import matplotlib.pyplot as plt  # this import is needed for interactive tests
import numpy as np
import pytest

from locan import (  # noqa: F401  # this import is needed for interactive tests
    RenderEngine,
    apply_window,
    cluster_dbscan,
    render_2d_mpl,
    render_2d_rgb_mpl,
    render_2d_scatter_density,
    scatter_2d_mpl,
    transform_affine,
)
from locan.dependencies import HAS_DEPENDENCY


def test_render_2d_mpl_empty(locdata_empty):
    render_2d_mpl(locdata_empty, bin_size=5)
    # plt.show()

    plt.close("all")


def test_render_2d_mpl_single(locdata_single_localization, caplog):
    render_2d_mpl(locdata_single_localization, bin_size=0.5)
    log_msg = (
        "locan.visualize.render_mpl.render2d",
        30,
        "Locdata carries a single localization.",
    )
    assert log_msg in caplog.record_tuples
    # plt.show()

    plt.close("all")


def test_render_2d_mpl(locdata_blobs_2d):
    render_2d_mpl(locdata_blobs_2d)
    render_2d_mpl(
        locdata_blobs_2d, bin_size=100, bin_range=[[500, 1000], [500, 1000]], cbar=False
    )

    render_2d_mpl(locdata_blobs_2d, bin_size=100, bin_range=None, rescale=None)
    render_2d_mpl(locdata_blobs_2d, bin_size=100, bin_range=None, rescale=True)
    render_2d_mpl(locdata_blobs_2d, bin_size=100, bin_range=None, rescale="unity")
    render_2d_mpl(locdata_blobs_2d, bin_size=100, bin_range=None, rescale="equal")
    render_2d_mpl(locdata_blobs_2d, bin_size=100, bin_range=None, rescale=(0, 50))
    render_2d_mpl(locdata_blobs_2d, bin_size=100, bin_range="zero")

    fig, ax = plt.subplots(nrows=1, ncols=2)
    render_2d_mpl(locdata_blobs_2d, ax=ax[0])
    render_2d_mpl(locdata_blobs_2d, bin_range="zero", ax=ax[1])

    render_2d_mpl(locdata_blobs_2d, ax=ax[0], colorbar_kws=dict(ax=ax[0]))
    render_2d_mpl(locdata_blobs_2d, bin_range="zero", ax=ax[1])

    plt.close("all")


@pytest.mark.visual
# this is to check overlay of rendered image and single localization points
def test_render_2d_mpl_show(locdata_blobs_2d):
    # print(locdata_blobs_2d.coordinates)
    render_2d_mpl(
        locdata_blobs_2d,
        bin_size=10,
        bin_range=None,
        rescale=None,
        other_property="position_y",
    )
    # plt.plot(locdata_blobs_2d.coordinates[:, 0],
    # locdata_blobs_2d.coordinates[:, 1],
    # 'o'
    # )
    plt.show()

    plt.close("all")


@pytest.mark.skipif(
    not HAS_DEPENDENCY["mpl_scatter_density"], reason="requires mpl_scatter_density"
)
def test_render_2d_scatter_density(locdata_blobs_2d):
    render_2d_scatter_density(locdata_blobs_2d)
    # render_2d_scatter_density(locdata_blobs_2d, bin_range=[[500, 1000], [500, 1000]],
    # cbar=False)

    # render_2d_scatter_density(locdata_blobs_2d, bin_range=None, vmin=0, vmax=1)

    # fig, ax = plt.subplots(nrows=1, ncols=2)
    # render_2d_scatter_density(locdata_blobs_2d, ax=ax[0])
    # render_2d_scatter_density(locdata_blobs_2d, bin_range='zero', ax=ax[1])
    #
    # render_2d_scatter_density(locdata_blobs_2d, ax=ax[0], colorbar_kws=dict(ax=ax[0]))
    # render_2d_scatter_density(locdata_blobs_2d, bin_range='zero', ax=ax[1])

    # render_2d_scatter_density(locdata_blobs_2d, other_property='position_x')

    # plt.show()

    plt.close("all")


@pytest.mark.skipif(
    not HAS_DEPENDENCY["mpl_scatter_density"], reason="requires mpl_scatter_density"
)
def test_render_2d_scatter_density_empty(locdata_empty):
    render_2d_scatter_density(locdata_empty)
    # plt.show()

    plt.close("all")


@pytest.mark.skipif(
    not HAS_DEPENDENCY["mpl_scatter_density"], reason="requires mpl_scatter_density"
)
def test_render_2d_scatter_density_single(locdata_single_localization, caplog):
    render_2d_scatter_density(locdata_single_localization)
    log_msg = (
        "locan.visualize.render_mpl.render2d",
        30,
        "Locdata carries a single localization.",
    )
    assert log_msg in caplog.record_tuples
    # plt.show()
    plt.close("all")


def test_scatter_2d_mpl(locdata_2d):
    scatter_2d_mpl(locdata_2d, text_kwargs=dict(color="r"), color="r")
    # plt.show()

    plt.close("all")


def test_scatter_2d_mpl_empty(locdata_empty):
    scatter_2d_mpl(locdata_empty)
    # plt.show()

    plt.close("all")


def test_scatter_2d_mpl_single(locdata_single_localization, caplog):
    scatter_2d_mpl(locdata_single_localization)
    log_msg = (
        "locan.visualize.render_mpl.render2d",
        30,
        "Locdata carries a single localization.",
    )
    assert log_msg in caplog.record_tuples
    # plt.show()

    plt.close("all")


@pytest.mark.visual  # Visual check repeating previously checked functionality
def test_scatter_2d_mpl_2(locdata_blobs_2d):
    _, collection = cluster_dbscan(locdata_blobs_2d, eps=20, min_samples=3)
    render_2d_mpl(locdata_blobs_2d)
    scatter_2d_mpl(collection)
    plt.show()

    plt.close("all")


def test_apply_window():
    img = np.ones((10, 10))
    img_filtered = apply_window(image=img, window_function="tukey", alpha=0.4)
    assert np.array_equal(img_filtered[0, :], np.zeros(10))
    assert np.array_equal(img_filtered[:, 0], np.zeros(10))


def test_render_2d_rgb_mpl_empty(locdata_empty):
    render_2d_rgb_mpl([locdata_empty, locdata_empty], bin_size=1)
    # plt.show()

    plt.close("all")


def test_render_2d_rgb_mpl_single(locdata_empty, locdata_single_localization, caplog):
    render_2d_rgb_mpl([locdata_empty, locdata_single_localization], bin_size=1)
    log_msg = (
        "locan.visualize.render_mpl.render2d",
        30,
        "Locdata carries a single localization.",
    )
    assert log_msg in caplog.record_tuples
    # plt.show()

    plt.close("all")


def test_render_2d_rgb_mpl(locdata_2d):
    render_2d_rgb_mpl([locdata_2d, locdata_2d], bin_size=1)
    # plt.show()

    plt.close("all")


@pytest.mark.visual  # Visual check repeating previously checked functionality
def test_render_2d_rgb_mpl_2(locdata_blobs_2d):
    locdata_0 = locdata_blobs_2d
    locdata_1 = transform_affine(locdata_blobs_2d, offset=(20, 0))
    render_2d_rgb_mpl([locdata_0, locdata_1], bin_size=20)
    plt.show()

    plt.close("all")


@pytest.mark.visual  # Visual check repeating previously checked functionality
def test_render_2d_rgb_mpl_3(locdata_blobs_2d):
    """Check intensity normalization."""
    locdata_0 = locdata_blobs_2d
    locdata_1 = transform_affine(locdata_blobs_2d, offset=(20, 0))

    render_2d_rgb_mpl([locdata_0, locdata_1], bin_size=20)
    plt.show()
    render_2d_rgb_mpl([locdata_0, locdata_1], bin_size=20, rescale=False)
    plt.show()

    render_2d_rgb_mpl([locdata_0, locdata_1], bin_size=100)
    plt.show()
    render_2d_rgb_mpl([locdata_0, locdata_1], bin_size=100, rescale=True)
    plt.show()
    render_2d_rgb_mpl([locdata_0, locdata_1], bin_size=100, rescale=False)
    plt.show()

    plt.close("all")



from locan.scripts import script_show_versions


def test_script_show_version(capfd):
    script_show_versions.main([])
    captured = capfd.readouterr()
    for header in ["\nLocan:", "\nPython:", "\nSystem:", "\nPython dependencies:"]:
        assert header in captured.out
    for header in ["node:", "executable:", "locan"]:
        assert header not in captured.out

    script_show_versions.main(["-o", "locan"])
    captured = capfd.readouterr()
    for header in [
        "\nLocan:",
        "\nPython:",
        "\nSystem:",
        "\nPython dependencies:",
        "locan",
    ]:
        assert header in captured.out

    script_show_versions.main(["-v", "-e", "-o", "locan"])
    captured = capfd.readouterr()
    for header in [
        "\nLocan:",
        "\nPython:",
        "\nSystem:",
        "\nPython dependencies:",
        "locan",
        "node:",
        "executable:",
    ]:
        assert header in captured.out



import pickle
import tempfile
from pathlib import Path

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_example import (
    AnalysisExampleAlgorithm_1,
    AnalysisExampleAlgorithm_2,
)


def test_Analysis_example():
    ae = AnalysisExampleAlgorithm_1(
        limits=(100, 110), meta={"comment": "this is an example"}
    )
    # print(ae)
    assert repr(ae) == "AnalysisExampleAlgorithm_1(limits=(100, 110))"
    assert str(ae) == repr(ae)
    assert ae.parameter == {"limits": (100, 110)}
    assert ae.meta.comment == "this is an example"
    ae.compute(locdata=None)
    assert len(ae.results) == 10

    ae.plot()
    ae.plot_2()
    ae.plot_histogram_fit()
    assert ae.a_center
    assert ae.a_sigma
    ae.report()

    ae_2 = eval(repr(ae))
    assert ae_2.parameter == {"limits": (100, 110)}


def test_Analysis_example_2():
    ae = AnalysisExampleAlgorithm_2()
    assert repr(ae) == "AnalysisExampleAlgorithm_2(n_sample=100, seed=None)"
    ae.compute()
    assert len(ae.results) == 100
    ae.plot()
    ae.plot_2()
    ae.plot_histogram_fit()
    assert ae.a_center
    assert ae.a_sigma
    ae.report()


def test_pickling_Analysis():
    ae = AnalysisExampleAlgorithm_1(
        limits=(100, 110), meta={"comment": "this is an example"}
    )
    ae.compute(locdata=None)
    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / "pickled_analysis.pickle"
        with open(file_path, "wb") as file:
            pickle.dump(ae, file, pickle.HIGHEST_PROTOCOL)
        with open(file_path, "rb") as file:
            ae_pickled = pickle.load(file)
        assert ae_pickled.parameter == ae.parameter
        assert isinstance(ae_pickled.meta, metadata_analysis_pb2.AMetadata)
        assert ae_pickled.meta == ae.meta



from copy import deepcopy

import matplotlib.pyplot as plt
import pandas as pd
import pytest

from locan import LocalizationsPerFrame, LocData
from locan.analysis.localizations_per_frame import (
    _DistributionFits,
    _localizations_per_frame,
    _Results,
)


def test__localizations_per_frame(caplog, locdata_rapidSTORM_2d):
    series = _localizations_per_frame(locdata=locdata_rapidSTORM_2d)
    assert len(series) == 48
    assert series.iloc[0] == 22
    assert series.index[-1] == 47
    assert series.name == "n_localizations"

    series = _localizations_per_frame(
        locdata=locdata_rapidSTORM_2d, norm=2, time_delta=None
    )
    assert len(series) == 48
    assert series.iloc[0] == 11
    assert series.index[-1] == 47
    assert series.name == "n_localizations / 2"

    series = _localizations_per_frame(locdata=locdata_rapidSTORM_2d, norm=2)
    assert len(series) == 48
    assert series.iloc[0] == 11
    assert series.index[-1] == 47

    series = _localizations_per_frame(
        locdata=locdata_rapidSTORM_2d, time_delta=pd.Timedelta(10, "ms")
    )
    assert len(series) == 48
    assert series.iloc[0] == 22
    assert series.index[-1].value == 47 * 10_000_000
    assert series.name == "n_localizations"

    series = _localizations_per_frame(locdata=locdata_rapidSTORM_2d, time_delta="10ms")
    assert len(series) == 48
    assert series.iloc[0] == 22
    assert series.index[-1].value == 47 * 10_000_000
    assert series.name == "n_localizations"

    series = _localizations_per_frame(
        locdata=locdata_rapidSTORM_2d, time_delta="10000us"
    )
    assert len(series) == 48
    assert series.iloc[0] == 22
    assert series.index[-1].value == 47 * 10_000_000
    assert series.name == "n_localizations"

    series = _localizations_per_frame(locdata=locdata_rapidSTORM_2d, time_delta=10)
    assert len(series) == 48
    assert series.iloc[0] == 22
    assert series.index[-1].value == 47 * 10_000_000
    assert series.name == "n_localizations"

    series = _localizations_per_frame(locdata=range(10))
    assert len(series) == 10
    assert series.iloc[0] == 1
    assert series.index[-1] == 9
    assert series.name == "n_localizations"

    series = _localizations_per_frame(
        locdata=locdata_rapidSTORM_2d, time_delta=10, resample="100ms"
    )
    assert len(series) == 5
    assert series.iloc[0] == 215
    assert series.index[-1].value == 400_000_000
    assert series.name == "n_localizations"

    with pytest.raises(TypeError):
        _localizations_per_frame(locdata=range(10), resample="2s")

    assert caplog.record_tuples == [
        (
            "locan.analysis.localizations_per_frame",
            30,
            "integration_time not available in locdata.meta - frames used instead.",
        ),
        (
            "locan.analysis.localizations_per_frame",
            30,
            "integration_time not available in locdata.meta - frames used instead.",
        ),
        (
            "locan.analysis.localizations_per_frame",
            30,
            "integration_time not available in locdata.meta - frames used instead.",
        ),
        (
            "locan.analysis.localizations_per_frame",
            30,
            "integration_time not available in locdata.meta - frames used instead.",
        ),
    ]

    locdata_with_meta = deepcopy(locdata_rapidSTORM_2d)
    setup = locdata_with_meta.meta.experiment.setups.add()
    ou = setup.optical_units.add()
    ou.detection.camera.integration_time.FromMilliseconds(10)

    series = _localizations_per_frame(
        locdata=locdata_with_meta, time_delta="integration_time"
    )
    assert len(series) == 48
    assert series.iloc[0] == 22
    assert series.index[-1].value == 47 * 10_000_000
    assert series.name == "n_localizations"


class TestLocalizationsPerFrame:
    def test_init(self, locdata_rapidSTORM_2d):
        lpf = LocalizationsPerFrame(meta={"comment": "this is an example"})
        assert (
            str(lpf)
            == "LocalizationsPerFrame(norm=None, time_delta=integration_time, resample=None)"
        )
        assert lpf.results is None
        assert lpf.meta.comment == "this is an example"

    def test_empty_locdata(self, caplog):
        lpf = LocalizationsPerFrame().compute(LocData())
        lpf.fit_distributions()
        lpf.plot()
        lpf.hist()
        # plt.show()

        assert caplog.record_tuples == [
            ("locan.analysis.localizations_per_frame", 30, "Locdata is empty."),
            (
                "locan.analysis.localizations_per_frame",
                30,
                "No results available to fit.",
            ),
        ]

        plt.close("all")

    def test_compute(self, locdata_rapidSTORM_2d):
        lpf = LocalizationsPerFrame().compute(locdata=locdata_rapidSTORM_2d)
        assert isinstance(lpf.results, _Results)
        assert isinstance(lpf.results.time_series, pd.Series)

        # print(lpf.results.time_series.head())
        assert len(lpf.results.time_series) == 48
        assert lpf.results.time_series.iloc[0] == 22
        assert lpf.results.time_series.index[-1] == 47

        assert lpf.results.accumulation_time() == 22
        assert lpf.results.accumulation_time(fraction=0.8) == 38

        lpf.plot()
        lpf.plot(window=10)
        lpf.hist()
        # plt.show()

        lpf.fit_distributions(floc=0)
        assert 0 in lpf.distribution_statistics.parameter_dict().values()

        plt.close("all")

    def test_resample(self, locdata_rapidSTORM_2d):
        lpf = LocalizationsPerFrame(
            norm="localization_density_bb", time_delta=1, resample="2s"
        ).compute(locdata=locdata_rapidSTORM_2d)
        assert (
            lpf.results.time_series.name == "n_localizations / localization_density_bb"
        )

        lpf.plot()
        lpf.plot(window=10)
        lpf.hist()
        # plt.show()

        plt.close("all")

    def test_norm(self, locdata_rapidSTORM_2d):
        lpf = LocalizationsPerFrame(norm=5).compute(locdata=locdata_rapidSTORM_2d)
        assert lpf.results.time_series.name == "n_localizations / 5"

        locdata_rapidSTORM_2d_ = deepcopy(locdata_rapidSTORM_2d)
        od = locdata_rapidSTORM_2d_.meta.experiment.setups.add().optical_units.add()
        od.detection.camera.integration_time.FromMilliseconds(10)
        lpf = LocalizationsPerFrame(norm=None).compute(locdata=locdata_rapidSTORM_2d_)
        assert lpf.results.time_series.name == "n_localizations"


@pytest.mark.visual
class TestLocalizationsPerFrameVisual:
    @pytest.fixture()
    def lpf(self, locdata_rapidSTORM_2d):
        return LocalizationsPerFrame().compute(locdata=locdata_rapidSTORM_2d)

    def test_plot(self, lpf):
        lpf.plot()
        plt.show()

        lpf.plot(cumulative=True)
        plt.show()

        lpf.plot(cumulative=True, normalize=True)
        plt.show()

        plt.close("all")

    def test_hist(self, lpf):
        lpf.hist(density=False)
        plt.show()

        lpf.hist(density=True, cumulative=True)
        plt.show()

        plt.close("all")


class TestDistributionFits:
    def test_fit(self, locdata_rapidSTORM_2d):
        lpf = LocalizationsPerFrame().compute(locdata=locdata_rapidSTORM_2d)
        distribution_statistics = _DistributionFits(lpf)
        # print(distribution_statistics)
        assert distribution_statistics.parameter_dict() == {}

        distribution_statistics.fit()
        assert list(distribution_statistics.parameter_dict().keys()) == [
            "n_localizations_center",
            "n_localizations_sigma",
        ]

        distribution_statistics.plot()
        # plt.show()

        plt.close("all")



import matplotlib.colors as mcolors

from locan import (
    DECODE_KEYS,
    ELYRA_KEYS,
    NANOIMAGER_KEYS,
    PROPERTY_KEYS,
    RAPIDSTORM_KEYS,
    SMAP_KEYS,
    SMLM_KEYS,
    THUNDERSTORM_KEYS,
    ColorMaps,
    FileType,
    HullType,
    PropertyDescription,
    PropertyKey,
    RenderEngine,
)
from locan.dependencies import HAS_DEPENDENCY


def test_PropertyDescription():
    prop = PropertyDescription(
        name="prop_name",
        type="integer",
        unit_SI="m",
        unit="nm",
        description="something",
    )
    assert (
        repr(prop)
        == "PropertyDescription(name='prop_name', type='integer', unit_SI='m', unit='nm', "
        "description='something')"
    )
    assert prop.name == "prop_name"


def test_PropertyKey():
    assert "index" in PropertyKey.__members__
    for element in PropertyKey:
        assert element.name == element.value.name

    assert all(
        key in PropertyKey.coordinate_labels()
        for key in [
            PropertyKey.position_x,
            PropertyKey.position_y,
            PropertyKey.position_z,
        ]
    )

    PropertyKey.position_x.value.unit = "nm"
    assert PropertyKey.position_x.value.unit == "nm"

    string_ = PropertyKey.summary()
    # print(string_)
    assert string_[:5] == "index"


def test_PROPERTY_KEYS():
    assert all(key in PropertyKey._member_names_ for key in PROPERTY_KEYS.keys())


def test_HullType():
    values = [item.value for item in HullType]
    assert all(
        value in values
        for value in [
            "bounding_box",
            "convex_hull",
            "oriented_bounding_box",
            "alpha_shape",
        ]
    )


def test_FileType():
    from locan.data.metadata_pb2 import File_type

    for ft_enum, ft_pb in zip(FileType, File_type.items()):
        assert ft_enum.name == ft_pb[0]
        assert ft_enum.value == ft_pb[1]


def test_RenderEngine():
    if not HAS_DEPENDENCY["napari"]:
        assert all(key in RenderEngine._member_names_ for key in ["MPL"])
    else:
        assert all(key in RenderEngine._member_names_ for key in ["MPL", "NAPARI"])


def test_ColorMaps():
    assert all([isinstance(item.value, mcolors.Colormap) for item in ColorMaps])


def test_keys_are_mapped_on_valid_property_keys():
    assert all(PropertyKey[value] for value in RAPIDSTORM_KEYS.values())
    assert all(PropertyKey[value] for value in ELYRA_KEYS.values())
    assert all(PropertyKey[value] for value in THUNDERSTORM_KEYS.values())
    assert all(PropertyKey[value] for value in NANOIMAGER_KEYS.values())
    assert all(PropertyKey[value] for value in SMLM_KEYS.values())
    assert all(PropertyKey[value] for value in DECODE_KEYS.values())
    assert all(PropertyKey[value] for value in SMAP_KEYS.values())



"""

Hull objects of localization data.

This module computes specific hulls for the bounding box, convex hull and
oriented bounding box and related properties for LocData objects.

"""
from __future__ import annotations

import numpy as np
import scipy.spatial as spat
from shapely.geometry import MultiPoint as shMultiPoint

from locan.data.region import Polygon, Rectangle

__all__ = ["BoundingBox", "ConvexHull", "OrientedBoundingBox"]


class BoundingBox:
    """
    Class with bounding box computed using numpy operations.

    Parameters
    ----------
    points : numpy.ndarray[float]
        Coordinates of input points. Array with shape (npoints, ndim).

    Attributes
    ----------
    hull : numpy.ndarray
        Array of point coordinates of shape (2, ndim) that represent
        [[min_coordinates], [max_coordinates]].
    dimension : int
        Spatial dimension of hull
    vertices : numpy.ndarray
        Coordinates of points that make up the hull.
        Array of shape (ndim, 2).
    width : numpy.ndarray[float]
        Array with differences between max and min for each coordinate.
    region_measure : float
        Hull measure, i.e. area or volume
    subregion_measure : float
        Measure of the sub-dimensional region, i.e. circumference or surface
    region : RoiRegion
        Convert the hull to a RoiRegion object.
    """

    def __init__(self, points):
        if np.size(points) == 0:
            self.dimension = 0
            self.hull = np.array([])
            self.width = np.zeros(self.dimension)
            self.region_measure = 0
            self.subregion_measure = 0
        elif len(points) < 2:
            self.dimension = np.shape(points)[1]
            self.hull = np.array([])
            self.width = np.zeros(self.dimension)
            self.region_measure = 0
            self.subregion_measure = 0
        else:
            self.dimension = np.shape(points)[1]
            self.hull = np.array([np.min(points, axis=0), np.max(points, axis=0)])
            self.width = np.diff(self.hull, axis=0).flatten()
            self.region_measure = np.prod(self.width)
            self.subregion_measure = np.sum(self.width) * 2

    @property
    def vertices(self):
        return self.hull.T

    @property
    def region(self):
        if self.dimension == 2:
            region_ = Rectangle(self.hull[0], self.width[0], self.width[1], 0)
            # region_ = RoiRegion(region_type='rectangle', region_specs=(self.hull[0], self.width[0], self.width[1], 0))
        else:
            raise NotImplementedError
        return region_


class _ConvexHullScipy:
    """
    Class with convex hull computed using the scipy.spatial.ConvexHull method.

    Parameters
    ----------
    points : numpy.ndarray
        Coordinates of input points. Array with shape (npoints, ndim).

    Attributes
    ----------
    hull : scipy.spatial.ConvexHull
        hull object from the corresponding algorithm
    dimension : int
        spatial dimension of hull
    vertices : numpy.ndarray
        Coordinates of points that make up the hull.
        Array of shape (ndim, 2).
    vertex_indices : numpy.ndarray
        Indices identifying a polygon of all points that make up the hull.
    points_on_boundary : int
        absolute number of points that are part of the convex hull.
    points_on_boundary_rel : float
        The number of points on the hull relative to all input points
    region_measure : float
        hull measure, i.e. area or volume
    subregion_measure : float
        measure of the sub-dimensional region, i.e. circumference or surface
    region : Region
        Convert the hull to a Region object.
    """

    def __init__(self, points):
        if len(points) < 6:
            unique_points = np.array(list(set(tuple(point) for point in points)))
            if len(unique_points) < 3:
                raise TypeError(
                    "Convex_hull needs at least 3 different points as input."
                )

        self.dimension = np.shape(points)[1]
        self.hull = spat.ConvexHull(points)
        self.vertex_indices = self.hull.vertices
        self.points_on_boundary = len(self.vertex_indices)
        self.points_on_boundary_rel = self.points_on_boundary / len(points)
        self.region_measure = self.hull.volume
        self.subregion_measure = self.hull.area

    @property
    def vertices(self):
        return self.hull.points[self.hull.vertices]

    @property
    def region(self):
        if self.dimension > 2:
            raise NotImplementedError(
                "Region for 3D data has not yet been implemented."
            )
        else:
            # closed_vertices = np.append(self.vertices, [self.vertices[0]], axis=0)
            # region_ = RoiRegion(region_type='polygon', region_specs=closed_vertices)
            return Polygon(self.vertices)


class _ConvexHullShapely:
    """
    Class with convex hull computed using the scipy.spatial.ConvexHull method.

    Parameters
    ----------
    points : numpy.ndarray
        Coordinates of input points. Array with shape (npoints, ndim).

    Attributes
    ----------
    hull : Hull object
        Polygon object from the .convex_hull method
    dimension : int
        Spatial dimension of hull
    vertices : numpy.ndarray
        Coordinates of points that make up the hull.
        Array of shape (ndim, 2).
    vertex_indices : numpy.ndarray
        indices identifying a polygon of all points that make up the hull
    points_on_boundary : int
        The absolute number of points on the hull
    points_on_boundary_rel : int
        The number of points on the hull relative to all input points
    region_measure : float
        hull measure, i.e. area or volume
    subregion_measure : float
        measure of the sub-dimensional region, i.e. circumference or surface
    region : RoiRegion
        Convert the hull to a RoiRegion object.
    """

    def __init__(self, points):
        if len(points) < 6:
            unique_points = np.array(list(set(tuple(point) for point in points)))
            if len(unique_points) < 3:
                raise TypeError(
                    "Convex_hull needs at least 3 different points as input."
                )

        self.dimension = np.shape(points)[1]
        if self.dimension >= 3:
            raise TypeError(
                "ConvexHullShapely only takes 1 or 2-dimensional points as input."
            )

        self.hull = shMultiPoint(points).convex_hull
        # todo: set vertex_indices
        # self.vertex_indices = None
        self.points_on_boundary = (
            len(self.hull.exterior.coords) - 1
        )  # the first point is repeated in exterior.coords
        self.points_on_boundary_rel = self.points_on_boundary / len(points)
        self.region_measure = self.hull.area
        self.subregion_measure = self.hull.length

    @property
    def vertices(self):
        return np.array(self.hull.exterior.coords)[:-1]

    @property
    def region(self):
        if self.dimension > 2:
            raise NotImplementedError(
                "Region for 3D data has not yet been implemented."
            )
        else:
            #  closed_vertices = np.append(self.vertices, [self.vertices[0]], axis=0)
            # region_ = RoiRegion(region_type='polygon', region_specs=closed_vertices)
            return Polygon(self.vertices)


class ConvexHull:
    """
    Class with convex hull of localization data.

    Parameters
    ----------
    points : numpy.ndarray
        Coordinates of input points. Array with shape (npoints, ndim).
    method : str
        Specific class to compute the convex hull and attributes.
        One of 'scipy', 'shapely'.

    Attributes
    ----------
    method : str
        Specific class to compute the convex hull and attributes.
        One of 'scipy', 'shapely'.
    hull : Hull object
        Polygon object from the .convex_hull method
    dimension : int
        Spatial dimension of hull
    vertices : numpy.ndarray
        Coordinates of points that make up the hull.
        Array of shape (ndim, 2).
    vertex_indices : numpy.ndarray
        indices identifying a polygon of all points that make up the hull
    points_on_boundary : int
        The absolute number of points on the hull
    points_on_boundary_rel : int
        The number of points on the hull relative to all input points
    region_measure : float
        hull measure, i.e. area or volume
    subregion_measure : float
        measure of the sub-dimensional region, i.e. circumference or surface
    region : Region
        Convert the hull to a Region object.
    """

    def __init__(self, points, method="scipy"):
        self.method = method
        if method == "scipy":
            self._special_class = _ConvexHullScipy(points)
        elif method == "shapely":
            self._special_class = _ConvexHullShapely(points)
        else:
            raise ValueError(f"The provided method {method} is not available.")

    def __getattr__(self, attr):
        if attr.startswith("__") and attr.endswith(
            "__"
        ):  # this is needed to enable pickling
            raise AttributeError

        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self._special_class, attr)


class OrientedBoundingBox:
    """
    Class with oriented bounding box computed using the shapely
    minimum_rotated_rectangle method.

    Parameters
    ----------
    ----------
    points : numpy.ndarray
        Coordinates of input points. Array with shape (npoints, ndim).

    Attributes
    ----------
    hull : hull object
        Polygon object from the minimum_rotated_rectangle method
    dimension : int
        Spatial dimension of hull
    vertices : numpy.ndarray
        Coordinates of points that make up the hull.
        Array of shape (ndim, 2).
    width : numpy.ndarray[float]
        Array with lengths of box edges.
    region_measure : float
        hull measure, i.e. area or volume
    subregion_measure : float
        measure of the sub-dimensional region, i.e. circumference or surface
    region : Region
        Convert the hull to a Region object.
    angle : float
        Orientation defined as angle (in degrees) between the vector from
        first to last point and x-axis.

    """

    def __init__(self, points):
        self.dimension = np.shape(points)[1]

        if self.dimension >= 3:
            raise TypeError(
                "OrientedBoundingBox only takes 1 or 2-dimensional points as input."
            )

        if len(points) < 3:
            self.hull = np.array([])
            self.width = np.zeros(self.dimension)
            self.region_measure = 0
            self.subregion_measure = 0
            self.angle = np.nan
            self.elongation = np.nan
        else:
            self.hull = shMultiPoint(points).minimum_rotated_rectangle
            difference = np.diff(self.vertices[0:3], axis=0)
            self.width = np.array(
                [np.linalg.norm(difference[0]), np.linalg.norm(difference[1])]
            )
            self.region_measure = self.hull.area
            self.subregion_measure = self.hull.length
            self.angle = float(
                np.degrees(np.arctan2(difference[0][1], difference[0][0]))
            )
            # numpy.arctan2(y, x) takes reversed x, y arguments.
            self.elongation = 1 - np.divide(*sorted(self.width))

    @property
    def vertices(self):
        return np.array(self.hull.exterior.coords)

    @property
    def region(self):
        if self.dimension == 2:
            # region_ = RoiRegion(
            # region_type='rectangle',
            # region_specs=(self.vertices[0], self.width[0], self.width[1], self.angle)
            # )
            return Rectangle(self.vertices[0], self.width[0], self.width[1], self.angle)
        else:
            raise NotImplementedError



"""

Utility methods for clustering locdata.

"""
from __future__ import annotations

import itertools
import sys
from collections.abc import Callable  # noqa: F401

from locan.data.locdata import LocData

__all__ = ["serial_clustering"]


def serial_clustering(locdata, algorithm, parameter_lists, **kwargs):
    """
    Run and analyse a series of clustering processes to identify optimal
    parameters.

    Parameters
    ----------
    locdata : LocData
        Localization data.
    algorithm : Callable
        The locan clustering algorithm to use on locdata.
    parameter_lists : dict
        A dictionary with all parameter lists that are to be iterated.
        The keys should be identical to parameter names of the used algorithm.
    kwargs : dict
        Optional keyword arguments that are passed to the algorithm.

    Returns
    -------
    tuple[LocData]
        The first element is a LocData object with a selection of all
        localizations that are defined as noise.
        If noise is false this element will be None.
        The second element is a new LocData instance assembling all generated
        selections (i.e. localization cluster).
    """
    parameter = locals()

    keys = parameter_lists.keys()
    products = list(itertools.product(*parameter_lists.values()))
    products_transposed = zip(*products)
    dictionary = {key: values for key, values in zip(keys, products_transposed)}
    results = []
    for product in products:
        iterated_arguments = {k: v for k, v in zip(keys, product)}
        results.append(algorithm(locdata, **iterated_arguments, **kwargs))

    noise_locdata = [result[0] for result in results]
    noise_collection = LocData.from_collection(noise_locdata)
    noise_collection.dataframe = noise_collection.dataframe.assign(**dictionary)
    collection_locdata = [result[1] for result in results]
    collection = LocData.from_collection(collection_locdata)
    collection.dataframe = collection.dataframe.assign(**dictionary)

    # metadata for noise_collection
    del noise_collection.meta.history[:]
    noise_collection.meta.history.add(
        name=sys._getframe().f_code.co_name, parameter=str(parameter)
    )

    # metadata for collection
    del collection.meta.history[:]
    collection.meta.history.add(
        name=sys._getframe().f_code.co_name, parameter=str(parameter)
    )

    return noise_collection, collection



"""

Visualize localization data.

This module provides functions to visualize localization data
with matplotlib or napari.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   render
   render_mpl
   napari
   transform

"""

from .napari import *
from .render import *
from .render_mpl import *
from .transform import *

__all__ = []
__all__.extend(render.__all__)
__all__.extend(render_mpl.__all__)
__all__.extend(napari.__all__)
__all__.extend(transform.__all__)



"""

Compute statistics for localization data.

These values can represent new properties of locdata.

"""
from __future__ import annotations

from collections import namedtuple

import numpy as np
import pandas as pd

from locan.data.locdata import LocData

__all__ = ["statistics", "ranges", "range_from_collection"]


def statistics(
    locdata, statistic_keys=("count", "min", "max", "mean", "median", "std", "sem")
):
    """
    Compute selected statistical parameter for localization data.

    Parameters
    ----------
    locdata : LocData | pandas.DataFrame | pandas.Series
        Localization data

    statistic_keys : str | tuple[str]
        Pandas statistic functions.
        Default: ('count', 'min', 'max', 'mean', 'median', 'std', 'sem')

    Returns
    -------
    dict
        A dict with descriptive statistics.
    """

    if isinstance(locdata, LocData):
        data = locdata.data
    elif isinstance(locdata, (pd.DataFrame, pd.Series)):
        data = locdata
    else:
        raise TypeError("locdata should be of type locan.LocData or pandas.DataFrame.")

    statistics_ = data.agg(statistic_keys)

    if isinstance(locdata, pd.Series):
        p = str(data.name)
        if isinstance(statistic_keys, str):
            dict_ = {p + "_" + statistic_keys: statistics_}
        else:
            dict_ = {p + "_" + s: statistics_[s] for s in statistic_keys}
    else:
        if isinstance(statistic_keys, str):
            generator = (p for p in list(data))
            dict_ = {p + "_" + statistic_keys: statistics_[p] for p in generator}
        else:
            generator = ((p, s) for p in list(data) for s in statistic_keys)
            dict_ = {p + "_" + s: statistics_[p][s] for p, s in generator}

    return dict_


def ranges(locdata, loc_properties=None, special=None, epsilon=1):
    """
    Provide data ranges for localization properties.
    If LocData is empty, None is returned.
    If LocData carries a single localization, the range will be
    (value, value + `epsilon`).

    Parameters
    ----------
    locdata : LocData
        Localization data.
    loc_properties : str | tuple[str, ...] | list[str] | True | None
        Localization properties for which the range is determined.
        If None the ranges for all spatial coordinates are returned.
        If True the ranges for all localization properties are returned.
    special : str | None
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from
        all combined data are returned.
    epsilon : float
        number to specify the range for single values in locdata.

    Returns
    -------
    numpy.ndarray[float] | None
        The data range (min, max) for each localization property.
        Array of shape (dimension, 2).
    """
    if locdata.data.empty:
        return None
    elif len(locdata) == 1:
        pass

    if loc_properties is None:
        ranges_ = locdata.bounding_box.hull.T.copy()
    elif loc_properties is True:
        ranges_ = np.array([locdata.data.min(), locdata.data.max()]).T
    elif isinstance(loc_properties, str):
        ranges_ = np.array(
            [[locdata.data[loc_properties].min(), locdata.data[loc_properties].max()]]
        )
    else:
        loc_properties = list(loc_properties)
        ranges_ = np.array(
            [locdata.data[loc_properties].min(), locdata.data[loc_properties].max()]
        ).T

    if len(locdata) == 1:
        if ranges_.size == 0:
            ranges_ = np.concatenate(
                [locdata.coordinates, locdata.coordinates + epsilon], axis=0
            ).T
        else:
            ranges_ = ranges_ + [0, epsilon]

    if special is None:
        pass
    elif special == "zero":
        ranges_[:, 0] = 0
    elif special == "link":
        minmax = np.array([ranges_[:, 0].min(axis=0), ranges_[:, 1].max(axis=0)])
        ranges_ = np.repeat(minmax[None, :], len(ranges_), axis=0)
    else:
        raise ValueError(f"The parameter special={special} is not defined.")

    return ranges_


def range_from_collection(locdatas, loc_properties=None, special=None, epsilon=1):
    """
    Compute the maximum range from all combined localizations for each
    dimension.

    Parameters
    ----------
    locdatas : list[LocData]
        Collection of localization datasets.
    loc_properties : str | tuple[str, ...] | list[str] | True | None
        Localization properties for which the range is determined.
        If None the ranges for all spatial coordinates are returned.
        If True the ranges for all locdata.data properties are returned.
    special : str | None
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from
        all combined data are returned.
    epsilon : float
        number to specify the range for single values in locdata.

    Returns
    -------
    namedtuple
        A namedtuple('Ranges', locdata.coordinate_labels)
        of namedtuple('Range', 'min max').
    """
    ranges_ = [
        ranges(
            locdata=locdata,
            loc_properties=loc_properties,
            special=special,
            epsilon=epsilon,
        )
        for locdata in locdatas
    ]

    mins = np.min([rand[:, 0] for rand in ranges_], axis=0)
    maxs = np.max([rand[:, 1] for rand in ranges_], axis=0)

    if loc_properties is None:
        labels = locdatas[0].coordinate_labels
    else:
        labels = loc_properties

    Ranges = namedtuple("Ranges", labels)
    Range = namedtuple("Range", "min max")
    result = Ranges(
        *(Range(min_value, max_value) for min_value, max_value in zip(mins, maxs))
    )
    return result



from pathlib import Path

import matplotlib.colors as mcolors

from locan.configuration import (
    COLORMAP_CATEGORICAL,
    COLORMAP_CONTINUOUS,
    COLORMAP_DIVERGING,
    DATASETS_DIR,
    N_JOBS,
    QT_BINDING,
    RENDER_ENGINE,
)
from locan.constants import RenderEngine
from locan.dependencies import HAS_DEPENDENCY


def test_datasets_dir():
    # print(DATASETS_DIR)
    assert isinstance(DATASETS_DIR, Path)


def test_render_engine():
    assert isinstance(RENDER_ENGINE, RenderEngine)


def test_n_jobs():
    assert isinstance(N_JOBS, int)


def test_colormaps():
    colormaps = [COLORMAP_CONTINUOUS, COLORMAP_DIVERGING, COLORMAP_CATEGORICAL]
    for item in colormaps:
        assert isinstance(item, mcolors.Colormap) or isinstance(
            mcolors.Colormap(item), mcolors.Colormap
        )


def test_qt_bindings():
    assert "qt" in HAS_DEPENDENCY
    assert isinstance(QT_BINDING, str)

    if HAS_DEPENDENCY["qt"]:
        from qtpy import QT_VERSION

        assert QT_VERSION is not None



import pickle
import tempfile
from pathlib import Path

import pytest
from scipy import stats

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_base import _Analysis, _list_parameters


def test_Analysis():
    ae = _Analysis(limits=(100, 110), meta={"comment": "this is an example"})
    # print(repr(ae))
    assert str(ae) == repr(ae)
    assert ae.parameter == {"limits": (100, 110)}

    # print(ae.meta)
    # print(dir(metadata_analysis_pb2))
    assert isinstance(ae.meta, metadata_analysis_pb2.AMetadata)
    assert ae.meta.identifier
    assert ae.meta.creation_time
    assert ae.meta.method.name == "_Analysis"
    assert ae.meta.comment == "this is an example"

    ae.meta.comment = "new comment"
    assert ae.meta.comment == "new comment"

    ae.meta.map["variable key"] = "value_1"
    assert str(ae.meta.map) == "{'variable key': 'value_1'}"

    ae.meta.map["key_2"] = "value_2"
    assert ae.meta.map["key_2"] == "value_2"

    with pytest.raises(NotImplementedError):
        ae.compute()


def test_pickling_Analysis():
    ae = _Analysis(limits=(100, 110), meta={"comment": "this is an example"})
    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / "pickled_analysis.pickle"
        with open(file_path, "wb") as file:
            pickle.dump(ae, file, pickle.HIGHEST_PROTOCOL)
        with open(file_path, "rb") as file:
            ae_pickled = pickle.load(file)
        assert ae_pickled.parameter == ae.parameter
        assert isinstance(ae_pickled.meta, metadata_analysis_pb2.AMetadata)
        assert ae_pickled.meta == ae.meta


def test__list_parameters():
    with pytest.raises(AttributeError):
        parameters = _list_parameters("fail")
    parameters = _list_parameters("expon")
    assert parameters == ["loc", "scale"]
    parameters = _list_parameters(stats.gamma)
    assert parameters == ["a", "loc", "scale"]
    parameters = _list_parameters(stats.poisson)
    assert parameters == ["mu", "loc"]



"""

This module provides functions for rendering `LocData` objects in 3D.

"""
from __future__ import annotations

import logging

from matplotlib import pyplot as plt

from locan.data import LocData  # noqa: F401 # for typing

__all__ = ["scatter_3d_mpl"]

logger = logging.getLogger(__name__)


def scatter_3d_mpl(locdata, ax=None, index=True, text_kwargs=None, **kwargs):
    """
    Scatter plot of locdata elements with text marker for each element.

    Parameters
    ----------
    locdata : LocData
       Localization data.
    ax : matplotlib.axes.Axes3D
       The axes on which to show the plot
    index : bool
       Flag indicating if element indices are shown.
    text_kwargs : dict
       Keyword arguments for :func:`matplotlib.axes.Axes.text`.
    kwargs : dict
       Other parameters passed to :func:`matplotlib.axes.Axes.scatter`.

    Returns
    -------
    matplotlib.axes.Axes
       Axes object with the image.
    """
    if text_kwargs is None:
        text_kwargs = {}

    # Provide matplotlib.axes.Axes if not provided
    if ax is None:
        ax = plt.gca()

    # return ax if no or single point in locdata
    if len(locdata) < 2:
        if len(locdata) == 1:
            logger.warning("Locdata carries a single localization.")
        return ax

    coordinates = locdata.coordinates
    ax.scatter(*coordinates.T, **dict({"marker": "+", "color": "grey"}, **kwargs))

    # plot element number
    if index:
        for centroid, marker in zip(coordinates, locdata.data.index.values):
            ax.text(
                *centroid, marker, **dict({"color": "grey", "size": 20}, **text_kwargs)
            )

    ax.set(xlabel="position_x", ylabel="position_y", zlabel="position_z")

    return ax



import logging
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import locan.data.metadata_pb2
from locan import ROOT_DIR, LocData
from locan.dependencies import HAS_DEPENDENCY
from locan.locan_io.locdata.io_locdata import load_txt_file
from locan.locan_io.locdata.rapidstorm_io import load_rapidSTORM_file

logger = logging.getLogger(__name__)

for key, value in HAS_DEPENDENCY.items():
    if not value:
        logger.info(f"Extra dependency {key} is not available.")


# register pytest markers - should be in sync with pyproject.toml
def pytest_configure(config):
    config.addinivalue_line(
        "markers", "gui: marks tests that require user interaction (skipped by default)"
    )
    config.addinivalue_line(
        "markers",
        "visual: marks tests that require visual inspection (skipped by default)",
    )
    config.addinivalue_line(
        "markers",
        "requires_datasets: marks tests that require data in locan_datasets directory (skipped by default)",
    )


# fixtures for random points


@pytest.fixture(scope="session")
def few_random_points():
    """
    Fixture for returning few 2D points.
    """
    points = np.array(
        [
            [0.066, 0.64],
            [0.92, 0.65],
            [0.11, 0.40],
            [0.20, 0.17],
            [0.75, 0.92],
            [0.01, 0.12],
            [0.23, 0.54],
            [0.05, 0.25],
            [0.70, 0.73],
            [0.43, 0.16],
        ]
    )
    return points


# fixtures for LocData objects


@pytest.fixture(scope="session")
def locdata_empty():
    """
    Fixture for returning empty `LocData`.
    """
    df = pd.DataFrame()
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope="session")
def locdata_single_localization():
    """
    Fixture for returning `LocData` carrying a single 2D localization.
    """
    locdata_dict = {
        "position_x": np.array([1]),
        "position_y": np.array([1]),
        "frame": np.array([1]),
        "intensity": np.array([1]),
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope="session")
def locdata_single_localization_3d():
    """
    Fixture for returning `LocData` carrying a single 3D localization.
    """
    locdata_dict = {
        "position_x": np.array([1]),
        "position_y": np.array([1]),
        "position_z": np.array([1]),
        "frame": np.array([1]),
        "intensity": np.array([1]),
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope="session")
def locdata_1d():
    """
    Fixture for returning `LocData` carrying 1D localizations.
    """
    locdata_dict = {
        "position_x": np.array([1, 1, 2, 3, 4, 5]),
        "frame": np.array([1, 2, 2, 4, 5, 6]),
        "intensity": np.array([100, 150, 110, 80, 105, 95]),
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope="session")
def locdata_2d():
    """
    Fixture for returning `LocData` carrying 2D localizations.
    """
    locdata_dict = {
        "position_x": np.array([1, 1, 2, 3, 4, 5]),
        "position_y": np.array([1, 5, 3, 6, 2, 5]),
        "frame": np.array([1, 2, 2, 4, 5, 6]),
        "intensity": np.array([100, 150, 110, 80, 105, 95]),
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope="session")
def locdata_2d_negative():
    """
    Fixture for returning `LocData` carrying 2D localizations including
    negative coordinates.
    """
    locdata_dict = {
        "position_x": np.array([1, -1, 2, 3, 4, 5]),
        "position_y": np.array([1, 5, 3, 6, -2, 5]),
        "frame": np.array([1, 2, 2, 4, 5, 6]),
        "intensity": np.array([100, 150, 110, 80, 105, 95]),
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope="session")
def locdata_3d():
    """
    Fixture for returning `LocData` carrying 3D localizations.
    """
    locdata_dict = {
        "position_x": np.array([1, 1, 2, 3, 4, 5]),
        "position_y": np.array([1, 5, 3, 6, 2, 5]),
        "position_z": np.array([1, 2, 5, 4, 3, 2]),
        "frame": np.array([1, 2, 2, 4, 5, 6]),
        "intensity": np.array([100, 150, 110, 80, 105, 95]),
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope="session")
def locdata_non_standard_index():
    """
    Fixture for returning `LocData` carrying 2D localizations with arbitrary
    index.
    """
    locdata_dict = {
        "position_x": np.array([1, 1, 2, 3, 4, 5]),
        "position_y": np.array([1, 5, 3, 6, 2, 5]),
        "frame": np.array([1, 2, 2, 4, 5, 6]),
        "intensity": np.array([100, 150, 110, 80, 105, 95]),
    }
    df = pd.DataFrame(locdata_dict)
    df.index = [2, 1, 5, 10, 100, 200]
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    return LocData.from_dataframe(dataframe=df, meta=meta_)


@pytest.fixture(scope="session")
def locdata_rapidSTORM_2d():
    """
    Fixture for returning `LocData` carrying 2D localizations from
    rapidSTORM_dstorm_data.txt.
    """
    path = Path(ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_data.txt")
    dat = load_rapidSTORM_file(path)
    dat.meta.creation_time.FromSeconds(1)
    return dat


@pytest.fixture(scope="session")
def locdata_blobs_2d():
    """
    Fixture for returning `LocData` carrying 2D localizations from
    five_blobs.txt.
    """
    path = Path(ROOT_DIR / "tests/test_data/five_blobs.txt")
    dat = load_txt_file(path)
    dat.meta.creation_time.FromSeconds(1)
    return dat


@pytest.fixture(scope="session")
def locdata_blobs_3d():
    """
    Fixture for returning `LocData` carrying 3D localizations from
    five_blobs.txt.
    """
    path = Path(ROOT_DIR / "tests/test_data/five_blobs_3D.txt")
    dat = load_txt_file(path)
    dat.meta.creation_time.FromSeconds(1)
    return dat


@pytest.fixture(scope="session")
def locdata_two_cluster_2d():
    """
    Fixture for returning `LocData` carrying 2D localizations grouped in two
    clusters as indicated by `cluster_label`.
    """
    points = np.array([[0.5, 0.5], [1, 0.6], [1.1, 1], [5, 5.6], [5.1, 6], [5.5, 5]])
    locdata_dict = {
        "position_x": points.T[0],
        "position_y": points.T[1],
        "cluster_label": np.array([1, 1, 1, 2, 2, 2]),
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    locdata = LocData.from_dataframe(dataframe=df, meta=meta_)
    locdata.region = locdata.bounding_box.region
    return locdata


@pytest.fixture(scope="session")
def locdata_two_cluster_with_noise_2d():
    """
    Fixture for returning `LocData` carrying 2D localizations grouped in two
    clusters as indicated by `cluster_label` and noise.
    """
    points = np.array(
        [[0.5, 0.5], [1, 0.6], [1.1, 1], [5, 5.6], [5.1, 6], [5.5, 5], [100, 100]]
    )
    locdata_dict = {
        "position_x": points.T[0],
        "position_y": points.T[1],
        "cluster_label": np.array([1, 1, 1, 2, 2, 2, -1]),
    }
    df = pd.DataFrame(locdata_dict)
    meta_ = locan.data.metadata_pb2.Metadata()
    meta_.creation_time.seconds = 1
    locdata = LocData.from_dataframe(dataframe=df, meta=meta_)
    locdata.region = locdata.bounding_box.region
    return locdata



"""

This module provides functions to interact with napari
for rendering locdata objects in 2D.

"""
from __future__ import annotations

import logging

import numpy as np
from matplotlib import colors as mcolors

from locan import locdata_id
from locan.data import LocData
from locan.data.aggregate import Bins, _check_loc_properties, histogram
from locan.dependencies import HAS_DEPENDENCY
from locan.visualize.transform import adjust_contrast

if HAS_DEPENDENCY["napari"]:
    import napari

__all__ = [
    "render_2d_napari",
    "render_2d_rgb_napari",
]

logger = logging.getLogger(__name__)


def render_2d_napari(
    locdata,
    loc_properties=None,
    other_property=None,
    bins=None,
    n_bins=None,
    bin_size=None,
    bin_edges=None,
    bin_range=None,
    rescale=None,
    viewer=None,
    cmap="viridis",
    **kwargs,
) -> napari.Viewer:
    """
    Render localization data into a 2D image by binning x,y-coordinates into
    regular bins. Render the data using napari.

    Parameters
    ----------
    locdata : LocData
        Localization data.
    loc_properties : list[str] | None
        Localization properties to be grouped into bins.
        If None The coordinate_values of `locdata` are used.
    other_property : str | None
        Localization property (columns in `locdata.data`) that is averaged
        in each pixel.
        If None, localization counts are shown.
    bins : int | sequence | Bins | boost_histogram.axis.Axis | None
        The bin specification as defined in :class:`Bins`
    bin_edges : tuple | list | numpy.ndarray[float] | None
        Array of bin edges with shape (dimension, n_bin_edges)
        for all or each dimension.
    n_bins : int | list[int] | tuple[int] | numpy.ndarray[int] | None
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size : float | list[float] | tuple[float] | numpy.ndarray[float] | None
        The size of bins in units of `locdata` coordinate units for all
        or each dimension.
        5 would describe bin_size of 5 for all bins in all dimensions.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other dimension.
        To specify arbitrary sequence of `bin_sizes` use `bin_edges` instead.
    bin_range : tuple | tuple[tuples[float]] | str | None
        The data bin_range with shape (dimension, 2) to be taken into
        consideration for all or each dimension.
        ((min_x, max_x), (min_y, max_y), ...) bin_range for each coordinate;
        for None (min, max) bin_range are determined from data;
        for 'zero' (0, max) bin_range with max determined from data.
    rescale : int | str | locan.Trafo | callable | bool | None
        Transformation as defined in :class:`locan.Trafo` or by
        transformation function.
        For None or False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min intensities
        For 'equal' intensity values are rescaled by histogram equalization.
    viewer : napari.Viewer
        The viewer object on which to add the image
    cmap : str | Colormap
        The Colormap object used to map normalized data values to RGBA colors.
    kwargs : dict
        Other parameters passed to :func:`napari.Viewer.add_image`.

    Returns
    -------
    napari.Viewer
    """
    if not HAS_DEPENDENCY["napari"]:
        raise ImportError("Function requires napari.")

    # Provide napari viewer if not provided
    if viewer is None:
        viewer = napari.Viewer()

    # return if no or single point in locdata
    if len(locdata) < 2:
        if len(locdata) == 1:
            logger.warning("Locdata carries a single localization.")
        return viewer

    data, bins, labels = histogram(
        locdata=locdata,
        loc_properties=loc_properties,
        other_property=other_property,
        bins=bins,
        n_bins=n_bins,
        bin_size=bin_size,
        bin_edges=bin_edges,
        bin_range=bin_range,
    )
    if not all(bins.is_equally_sized):
        raise ValueError("All bins must be equally sized.")
    data = adjust_contrast(data, rescale)

    add_image_kwargs = dict(
        name=f"LocData {locdata_id}",
        colormap=cmap,
        scale=bins.bin_size,
        translate=np.asarray(bins.bin_range)[:, 0],
        metadata=dict(message=locdata.meta.SerializeToString()),
    )
    viewer.add_image(data=data, **dict(add_image_kwargs, **kwargs))

    # set scale_bar unit
    names = list(
        set(
            item.unit
            for item in locdata.meta.localization_properties
            if item.name in locdata.coordinate_labels
        )
    )
    if len(names) == 1:
        viewer.scale_bar.unit = f"1 {names[0]}"
    else:
        viewer.scale_bar.unit = None

    return viewer


def render_2d_rgb_napari(
    locdatas,
    loc_properties=None,
    other_property=None,
    bins=None,
    n_bins=None,
    bin_size=10,
    bin_edges=None,
    bin_range=None,
    rescale=None,
    viewer=None,
    **kwargs,
):
    """
    Render localization data into a 2D RGB image by binning x,y-coordinates into regular bins.

    Note
    ----
    For rescale=False no normalization is carried out image intensities are clipped to (0, 1) for float value
    or (0, 255) for integer values according to the matplotlib.imshow behavior.
    For rescale=None we apply a normalization to (min, max) of all intensity values.
    For all other rescale options the normalization is applied to each individual image.

    Parameters
    ----------
    locdatas : list[LocData]
        Localization data.
    loc_properties : list, None
        Localization properties to be grouped into bins. If None The coordinate_values of `locdata` are used.
    other_property : str, None
        Localization property (columns in `locdata.data`) that is averaged in each pixel.
        If None, localization counts are shown.
    bins : int, sequence, Bins, boost_histogram.axis.Axis, None
        The bin specification as defined in :class:`Bins`
    bin_edges : tuple, list, numpy.ndarray[float] with shape (dimension, n_bin_edges), None
        Array of bin edges for all or each dimension.
    n_bins : int, list, tuple, numpy.ndarray, None
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size : float, list, tuple, numpy.ndarray, None
        The size of bins in units of `locdata` coordinate units for all or each dimension.
        5 would describe bin_size of 5 for all bins in all dimensions.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other dimension.
        To specify arbitrary sequence of `bin_sizes` use `bin_edges` instead.
    bin_range : tuple, tuple[tuples[float]] with shape (dimension, 2), str, None
        The data bin_range to be taken into consideration for all or each dimension.
        ((min_x, max_x), (min_y, max_y), ...) bin_range for each coordinate;
        for None (min, max) bin_range are determined from data;
        for 'zero' (0, max) bin_range with max determined from data.
    rescale : int, str, locan.Trafo, callable, bool, None
        Transformation as defined in :class:`locan.Trafo` or by transformation function.
        For False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min intensities
        For 'equal' intensity values are rescaled by histogram equalization.
    viewer : napari.Viewer
        The viewer object on which to add the image
    cmap : str, Colormap
        The Colormap object used to map normalized data values to RGBA colors.
    kwargs : dict
        Other parameters passed to :func:`napari.Viewer.add_image`.

    Returns
    -------
    napari.Viewer, Bins
    """
    if not HAS_DEPENDENCY["napari"]:
        raise ImportError("Function requires napari.")

    # Provide napari viewer if not provided
    if viewer is None:
        viewer = napari.Viewer()

    locdata_temp = LocData.concat(locdatas)

    # return viewer if no or single point in locdata
    if len(locdata_temp) < 2:
        if len(locdata_temp) == 1:
            logger.warning("Locdata carries a single localization.")
        return viewer

    if bin_edges is None:
        _, bins, labels = histogram(
            locdata=locdata_temp,
            loc_properties=loc_properties,
            other_property=other_property,
            bins=bins,
            n_bins=n_bins,
            bin_size=bin_size,
            bin_edges=bin_edges,
            bin_range=bin_range,
        )

    else:
        labels = _check_loc_properties(locdata_temp, loc_properties)
        bins = Bins(bin_edges=bin_edges, labels=labels)

    imgs = [
        histogram(
            locdata=locdata,
            loc_properties=loc_properties,
            other_property=other_property,
            bin_edges=bins.bin_edges,
        ).data
        for locdata in locdatas
    ]

    if rescale is None:
        norm = mcolors.Normalize(vmin=np.min(imgs), vmax=np.max(imgs))
    else:
        norm = rescale
    imgs = [adjust_contrast(img, rescale=norm) for img in imgs]

    new = np.zeros_like(imgs[0])
    rgb_stack = np.stack([new] * 3, axis=2)

    for i, img in enumerate(imgs):
        rgb_stack[:, :, i] = img

    rgb_stack = np.transpose(rgb_stack, axes=(1, 0, 2))
    viewer.add_image(
        rgb_stack,
        **dict({"name": f"LocData {locdata_id}", "rgb": True}, **kwargs),
    )
    return viewer



import math

import matplotlib.pyplot as plt  # needed for visualization
import matplotlib.tri as mtri  # needed for visualization
import networkx as nx
import numpy as np
import pytest

from locan import AlphaComplex, AlphaShape, EmptyRegion, Region2D


@pytest.mark.visual
def test_AlphaComplex_visual(locdata_2d):
    points = locdata_2d.coordinates
    alpha_complex = AlphaComplex(points)

    alpha = 1.6
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type="all")
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(
        alpha, type="exterior"
    )
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(
        alpha, type="interior"
    )
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type="regular")
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(
        alpha, type="singular"
    )

    lengths = [
        len(sim)
        for sim in [
            ac_simplices_all,
            ac_simplices_exterior,
            ac_simplices_interior,
            ac_simplices_regular,
            ac_simplices_singular,
        ]
    ]
    assert lengths == [8, 2, 0, 6, 2]

    locdata_2d.data.plot(*locdata_2d.coordinate_labels, kind="scatter")
    for simp in ac_simplices_all:
        plt.plot(points[simp, 0], points[simp, 1], "-b")
    for simp in ac_simplices_interior:
        plt.plot(points[simp, 0], points[simp, 1], "--g")
    for simp in ac_simplices_regular:
        plt.plot(points[simp, 0], points[simp, 1], "--r")
    for simp in ac_simplices_singular:
        plt.plot(points[simp, 0], points[simp, 1], "--y")
    plt.show()


@pytest.mark.visual
def test_AlphaComplex_1_visual():
    points = np.array(
        [
            [10, 10],
            [10, 14],
            [11, 11],
            [12, 11],
            [13, 10],
            [13, 13],
            [15, 14],
            [17, 16],
            [10, 24],
        ]
    )

    alpha_complex = AlphaComplex(points)

    alpha = 2.2
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type="all")
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(  # noqa: F841
        alpha, type="exterior"
    )
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(
        alpha, type="interior"
    )
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type="regular")
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(
        alpha, type="singular"
    )

    H = alpha_complex.graph_from_lines(alpha, type="regular")

    fig, ax = plt.subplots(1, 3)
    nx.draw_networkx(H, points, with_labels=True, ax=ax[0])
    # plt.show()

    plt.scatter(*points.T)
    for simp in ac_simplices_all:
        ax[1].plot(points[simp, 0], points[simp, 1], "-b")
    for simp in ac_simplices_interior:
        ax[1].plot(points[simp, 0], points[simp, 1], "--g")
    for simp in ac_simplices_regular:
        ax[1].plot(points[simp, 0], points[simp, 1], "--r")
    for simp in ac_simplices_singular:
        ax[1].plot(points[simp, 0], points[simp, 1], "--y")
    # plt.show()

    triangles_all = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="all")
    ]
    triangles_exterior = alpha_complex.delaunay_triangulation.simplices[  # noqa: F841
        alpha_complex.get_alpha_complex_triangles(alpha, type="exterior")
    ]
    triangles_interior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="interior")
    ]
    triangles_regular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="regular")
    ]
    triangles_singular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="singular")
    ]

    tri = mtri.Triangulation(*points.T, triangles=triangles_all)
    ax[2].triplot(tri, "-b")
    tri = mtri.Triangulation(*points.T, triangles=triangles_interior)
    ax[2].triplot(tri, "--g")
    tri = mtri.Triangulation(*points.T, triangles=triangles_regular)
    ax[2].triplot(tri, "--r")
    tri = mtri.Triangulation(*points.T, triangles=triangles_singular)
    ax[2].triplot(tri, "--y")
    plt.show()


@pytest.mark.visual
def test_AlphaComplex_2_visual():
    points = np.array(
        [
            [10, 10],
            [10, 14],
            [11, 11],
            [12, 11],
            [13, 10],
            [13, 13],
            [15, 14],
            [17, 16],
            [20, 20],
            [20, 24],
            [21, 21],
            [22, 21],
            [23, 20],
            [23, 23],
            [25, 24],
            [10, 24],
        ]
    )

    alpha_complex = AlphaComplex(points)

    alpha = 2.2
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type="all")
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(  # noqa: F841
        alpha, type="exterior"
    )
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(
        alpha, type="interior"
    )
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type="regular")
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(
        alpha, type="singular"
    )

    H = alpha_complex.graph_from_lines(alpha, type="regular")

    fig, ax = plt.subplots(1, 3)
    nx.draw_networkx(H, points, with_labels=True, ax=ax[0])
    # plt.show()

    plt.scatter(*points.T)
    for simp in ac_simplices_all:
        ax[1].plot(points[simp, 0], points[simp, 1], "-b")
    for simp in ac_simplices_interior:
        ax[1].plot(points[simp, 0], points[simp, 1], "--g")
    for simp in ac_simplices_regular:
        ax[1].plot(points[simp, 0], points[simp, 1], "--r")
    for simp in ac_simplices_singular:
        ax[1].plot(points[simp, 0], points[simp, 1], "--y")
    # plt.show()

    triangles_all = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="all")
    ]
    triangles_exterior = alpha_complex.delaunay_triangulation.simplices[  # noqa: F841
        alpha_complex.get_alpha_complex_triangles(alpha, type="exterior")
    ]
    triangles_interior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="interior")
    ]
    triangles_regular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="regular")
    ]
    triangles_singular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="singular")
    ]

    tri = mtri.Triangulation(*points.T, triangles=triangles_all)
    ax[2].triplot(tri, "-b")
    tri = mtri.Triangulation(*points.T, triangles=triangles_interior)
    ax[2].triplot(tri, "--g")
    tri = mtri.Triangulation(*points.T, triangles=triangles_regular)
    ax[2].triplot(tri, "--r")
    tri = mtri.Triangulation(*points.T, triangles=triangles_singular)
    ax[2].triplot(tri, "--y")
    plt.show()


@pytest.mark.visual
def test_AlphaComplex_3_visual():
    points = np.array(
        [
            [10, 10],
            [10, 20],
            [20, 20],
            [20, 10],
            [10, 15],
            [15, 20],
            [20, 15],
            [15, 10],
            [10, 12],
            [10, 18],
            [20, 18],
            [20, 12],
            [17, 20],
            [12, 10],
            [17, 10],
            [12, 20],
            [11, 11],
            [11, 19],
            [19, 19],
            [19, 11],
            [11, 15],
            [15, 19],
            [19, 15],
            [15, 11],
            [13, 11],
            [11.5, 12],
            [12, 18],
            [13, 19],
            [17, 19],
            [18, 18],
            [17, 11],
            [18, 12],
            [15, 15],
        ]
    )

    alpha_complex = AlphaComplex(points)

    alpha = 2.2
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type="all")
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(  # noqa: F841
        alpha, type="exterior"
    )
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(
        alpha, type="interior"
    )
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type="regular")
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(
        alpha, type="singular"
    )

    H = alpha_complex.graph_from_lines(alpha, type="regular")

    fig, ax = plt.subplots(1, 3)
    nx.draw_networkx(H, points, with_labels=True, ax=ax[0])
    # plt.show()

    plt.scatter(*points.T)
    for simp in ac_simplices_all:
        ax[1].plot(points[simp, 0], points[simp, 1], "-b")
    for simp in ac_simplices_interior:
        ax[1].plot(points[simp, 0], points[simp, 1], "--g")
    for simp in ac_simplices_regular:
        ax[1].plot(points[simp, 0], points[simp, 1], "--r")
    for simp in ac_simplices_singular:
        ax[1].plot(points[simp, 0], points[simp, 1], "--y")
    # plt.show()

    triangles_all = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="all")
    ]
    triangles_exterior = alpha_complex.delaunay_triangulation.simplices[  # noqa: F841
        alpha_complex.get_alpha_complex_triangles(alpha, type="exterior")
    ]
    triangles_interior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="interior")
    ]
    triangles_regular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="regular")
    ]
    triangles_singular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="singular")
    ]

    tri = mtri.Triangulation(*points.T, triangles=triangles_all)
    ax[2].triplot(tri, "-b")
    tri = mtri.Triangulation(*points.T, triangles=triangles_interior)
    ax[2].triplot(tri, "--g")
    tri = mtri.Triangulation(*points.T, triangles=triangles_regular)
    ax[2].triplot(tri, "--r")
    tri = mtri.Triangulation(*points.T, triangles=triangles_singular)
    ax[2].triplot(tri, "--y")
    plt.show()


def test_AlphaComplex_0():
    points = np.array([])
    assert np.size(points) == 0
    alpha_complex = AlphaComplex(points)
    assert len(alpha_complex.triangles) == 0
    assert len(alpha_complex.lines) == 0
    assert alpha_complex.dimension is None
    assert (
        alpha_complex.get_alpha_complex_lines(
            alpha=alpha_complex.optimal_alpha(), type="exterior"
        )
        == []
    )
    assert alpha_complex.optimal_alpha() is None
    assert alpha_complex.alphas().size == 0
    # assert isinstance(alpha_complex.alpha_shape(alpha=2.2), AlphaShape)

    alpha = 2.2
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type="all")
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(
        alpha, type="exterior"
    )
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(
        alpha, type="interior"
    )
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type="regular")
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(
        alpha, type="singular"
    )

    lengths = [
        len(sim)
        for sim in [
            ac_simplices_all,
            ac_simplices_exterior,
            ac_simplices_interior,
            ac_simplices_regular,
            ac_simplices_singular,
        ]
    ]
    assert lengths == [0, 0, 0, 0, 0]

    triangles_all = alpha_complex.get_alpha_complex_triangles(alpha, type="all")
    assert len(triangles_all) == 0

    H = alpha_complex.graph_from_lines(alpha, type="regular")
    assert isinstance(H, nx.Graph)
    H = alpha_complex.graph_from_triangles(alpha, type="regular")
    assert isinstance(H, nx.Graph)


def test_AlphaComplex_1():
    points = np.array(
        [
            [10, 10],
            [10, 14],
            [11, 11],
            [12, 11],
            [13, 10],
            [13, 13],
            [15, 14],
            [17, 16],
            [10, 24],
        ]
    )
    assert len(points) == 9
    alpha_complex = AlphaComplex(points)
    assert len(alpha_complex.triangles) == 11
    assert len(alpha_complex.lines) == 19
    assert alpha_complex.dimension == 2
    assert (
        alpha_complex.get_alpha_complex_lines(
            alpha=alpha_complex.optimal_alpha(), type="exterior"
        )
        == []
    )
    assert alpha_complex.optimal_alpha() in alpha_complex.alphas()
    assert np.max(alpha_complex.alphas()) == np.inf
    assert isinstance(alpha_complex.alpha_shape(alpha=2.2), AlphaShape)

    alpha = 2.2
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type="all")
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(
        alpha, type="exterior"
    )
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(
        alpha, type="interior"
    )
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type="regular")
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(
        alpha, type="singular"
    )

    lengths = [
        len(sim)
        for sim in [
            ac_simplices_all,
            ac_simplices_exterior,
            ac_simplices_interior,
            ac_simplices_regular,
            ac_simplices_singular,
        ]
    ]
    assert lengths == [13, 6, 5, 5, 3]

    triangles_all = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="all")
    ]
    triangles_exterior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="exterior")
    ]
    triangles_interior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="interior")
    ]
    triangles_regular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="regular")
    ]
    triangles_singular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="singular")
    ]

    lengths = [
        len(sim)
        for sim in [
            triangles_all,
            triangles_exterior,
            triangles_interior,
            triangles_regular,
            triangles_singular,
        ]
    ]
    assert lengths == [6, 5, 1, 4, 1]

    H = alpha_complex.graph_from_lines(alpha, type="regular")
    assert isinstance(H, nx.Graph)
    H = alpha_complex.graph_from_triangles(alpha, type="regular")
    assert isinstance(H, nx.Graph)

    alpha = 1
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type="all")
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(
        alpha, type="exterior"
    )
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(
        alpha, type="interior"
    )
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type="regular")
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(
        alpha, type="singular"
    )

    lengths = [
        len(sim)
        for sim in [
            ac_simplices_all,
            ac_simplices_exterior,
            ac_simplices_interior,
            ac_simplices_regular,
            ac_simplices_singular,
        ]
    ]
    assert lengths == [3, 16, 0, 0, 3]

    triangles_all = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="all")
    ]
    triangles_exterior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="exterior")
    ]
    triangles_interior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="interior")
    ]
    triangles_regular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="regular")
    ]
    triangles_singular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="singular")
    ]

    lengths = [
        len(sim)
        for sim in [
            triangles_all,
            triangles_exterior,
            triangles_interior,
            triangles_regular,
            triangles_singular,
        ]
    ]
    assert lengths == [0, 11, 0, 0, 0]

    H = alpha_complex.graph_from_lines(alpha, type="all")
    assert isinstance(H, nx.Graph)
    H = alpha_complex.graph_from_triangles(alpha, type="regular")
    assert isinstance(H, nx.Graph)


def test_AlphaComplex_2():
    points = np.array(
        [
            [10, 10],
            [10, 14],
            [11, 11],
            [12, 11],
            [13, 10],
            [13, 13],
            [15, 14],
            [17, 16],
            [20, 20],
            [20, 24],
            [21, 21],
            [22, 21],
            [23, 20],
            [23, 23],
            [25, 24],
            [10, 24],
        ]
    )
    assert len(points) == 16
    alpha_complex = AlphaComplex(points)
    assert len(alpha_complex.lines) == 38
    assert alpha_complex.dimension == 2
    assert (
        alpha_complex.get_alpha_complex_lines(
            alpha=alpha_complex.optimal_alpha(), type="exterior"
        )
        == []
    )
    assert alpha_complex.optimal_alpha() in alpha_complex.alphas()
    assert np.max(alpha_complex.alphas()) == np.inf
    assert isinstance(alpha_complex.alpha_shape(alpha=2.2), AlphaShape)

    alpha = 2.2
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type="all")
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(
        alpha, type="exterior"
    )
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(
        alpha, type="interior"
    )
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type="regular")
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(
        alpha, type="singular"
    )

    lengths = [
        len(sim)
        for sim in [
            ac_simplices_all,
            ac_simplices_exterior,
            ac_simplices_interior,
            ac_simplices_regular,
            ac_simplices_singular,
        ]
    ]
    assert lengths == [25, 13, 10, 10, 5]

    triangles_all = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="all")
    ]
    triangles_exterior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="exterior")
    ]
    triangles_interior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="interior")
    ]
    triangles_regular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="regular")
    ]
    triangles_singular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="singular")
    ]

    lengths = [
        len(sim)
        for sim in [
            triangles_all,
            triangles_exterior,
            triangles_interior,
            triangles_regular,
            triangles_singular,
        ]
    ]
    assert lengths == [12, 11, 3, 7, 2]

    H = alpha_complex.graph_from_lines(alpha, type="regular")
    assert isinstance(H, nx.Graph)
    H = alpha_complex.graph_from_triangles(alpha, type="regular")
    assert isinstance(H, nx.Graph)


def test_AlphaComplex_3():
    points = np.array(
        [
            [10, 10],
            [10, 20],
            [20, 20],
            [20, 10],
            [10, 15],
            [15, 20],
            [20, 15],
            [15, 10],
            [10, 12],
            [10, 18],
            [20, 18],
            [20, 12],
            [17, 20],
            [12, 10],
            [17, 10],
            [12, 20],
            [11, 11],
            [11, 19],
            [19, 19],
            [19, 11],
            [11, 15],
            [15, 19],
            [19, 15],
            [15, 11],
            [13, 11],
            [11.5, 12],
            [12, 18],
            [13, 19],
            [17, 19],
            [18, 18],
            [17, 11],
            [18, 12],
            [15, 15],
        ]
    )
    assert len(points) == 33
    alpha_complex = AlphaComplex(points)
    assert len(alpha_complex.lines) == 80
    assert alpha_complex.dimension == 2
    assert (
        alpha_complex.get_alpha_complex_lines(
            alpha=alpha_complex.optimal_alpha(), type="exterior"
        )
        == []
    )
    assert alpha_complex.optimal_alpha() in alpha_complex.alphas()
    assert np.max(alpha_complex.alphas()) == np.inf
    assert isinstance(alpha_complex.alpha_shape(alpha=2.2), AlphaShape)

    alpha = 2.2
    ac_simplices_all = alpha_complex.get_alpha_complex_lines(alpha, type="all")
    ac_simplices_exterior = alpha_complex.get_alpha_complex_lines(
        alpha, type="exterior"
    )
    ac_simplices_interior = alpha_complex.get_alpha_complex_lines(
        alpha, type="interior"
    )
    ac_simplices_regular = alpha_complex.get_alpha_complex_lines(alpha, type="regular")
    ac_simplices_singular = alpha_complex.get_alpha_complex_lines(
        alpha, type="singular"
    )

    lengths = [
        len(sim)
        for sim in [
            ac_simplices_all,
            ac_simplices_exterior,
            ac_simplices_interior,
            ac_simplices_regular,
            ac_simplices_singular,
        ]
    ]
    assert lengths == [75, 5, 40, 28, 7]

    triangles_all = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="all")
    ]
    triangles_exterior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="exterior")
    ]
    triangles_interior = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="interior")
    ]
    triangles_regular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="regular")
    ]
    triangles_singular = alpha_complex.delaunay_triangulation.simplices[
        alpha_complex.get_alpha_complex_triangles(alpha, type="singular")
    ]

    lengths = [
        len(sim)
        for sim in [
            triangles_all,
            triangles_exterior,
            triangles_interior,
            triangles_regular,
            triangles_singular,
        ]
    ]
    assert lengths == [39, 9, 8, 28, 3]

    H = alpha_complex.graph_from_lines(alpha=2.2, type="regular")
    assert isinstance(H, nx.Graph)
    H = alpha_complex.graph_from_triangles(alpha, type="regular")
    assert isinstance(H, nx.Graph)


def test_AlphaShape_0():
    points = np.array([])

    alpha = 2.2
    alpha_shape = AlphaShape(alpha, points=points)
    assert len(alpha_shape.alpha_shape) == 0
    assert isinstance(alpha_shape.alpha_shape, list)
    assert alpha_shape.dimension is None
    assert isinstance(alpha_shape.region, EmptyRegion)
    assert alpha_shape.region_measure == 0
    assert alpha_shape.subregion_measure == 0
    assert alpha_shape.n_points_alpha_shape == 0
    assert math.isnan(alpha_shape.n_points_alpha_shape_rel)
    assert alpha_shape.n_points_on_boundary == 0
    assert math.isnan(alpha_shape.n_points_on_boundary_rel)
    assert alpha_shape.vertex_indices == []
    assert np.size(alpha_shape.vertices) == 0
    assert np.size(alpha_shape.vertices_alpha_shape) == 0
    assert np.size(alpha_shape.vertex_alpha_shape_indices) == 0

    cc = alpha_shape.connected_components
    assert len(cc) == 0
    assert alpha_shape.vertices_connected_components_indices == []


def test_AlphaShape_1():
    points = np.array(
        [
            [10, 10],
            [10, 14],
            [11, 11],
            [12, 11],
            [13, 10],
            [13, 13],
            [15, 14],
            [17, 16],
            [10, 24],
        ]
    )
    assert len(points) == 9

    alpha = 2.2
    alpha_shape = AlphaShape(alpha, points=points)
    assert len(alpha_shape.alpha_shape) == 13
    assert isinstance(alpha_shape.alpha_shape, list)
    assert alpha_shape.dimension == 2
    assert isinstance(alpha_shape.region, Region2D)
    assert alpha_shape.region_measure == 10.5
    assert alpha_shape.subregion_measure == 13.16227766016838
    assert alpha_shape.n_points_alpha_shape == 8
    assert alpha_shape.n_points_alpha_shape_rel == 0.8888888888888888
    assert alpha_shape.n_points_on_boundary == 5
    assert alpha_shape.n_points_on_boundary_rel == 0.625
    assert np.array_equal(alpha_shape.vertex_indices, [0, 1, 2, 4, 5])
    assert alpha_shape.vertices.shape == (5, 2)
    assert alpha_shape.vertices_alpha_shape.shape == (8, 2)
    assert len(alpha_shape.vertex_alpha_shape_indices) == 8
    cc = alpha_shape.connected_components
    assert len(cc) == 1
    assert isinstance(cc[0], Region2D)
    assert alpha_shape.vertices_connected_components_indices == [[2, 1, 0, 5, 3, 4]]

    alpha = 1
    # alpha_shape = AlphaShape(alpha, points=points)
    alpha_shape.alpha = 1
    assert len(alpha_shape.alpha_shape) == 3
    assert isinstance(alpha_shape.alpha_shape, list)
    assert alpha_shape.dimension == 2
    assert isinstance(alpha_shape.region, EmptyRegion)
    assert alpha_shape.region_measure == 0
    assert alpha_shape.subregion_measure == 0
    assert alpha_shape.n_points_alpha_shape == 4
    assert alpha_shape.n_points_alpha_shape_rel == 0.4444444444444444
    assert alpha_shape.n_points_on_boundary == 0
    assert alpha_shape.n_points_on_boundary_rel == 0
    assert np.array_equal(alpha_shape.vertex_indices, [])
    assert alpha_shape.vertices.shape == (0, 2)
    assert alpha_shape.vertices_alpha_shape.shape == (4, 2)
    assert len(alpha_shape.vertex_alpha_shape_indices) == 4
    cc = alpha_shape.connected_components
    assert len(cc) == 0
    assert alpha_shape.vertices_connected_components_indices == []


@pytest.mark.visual
def test_AlphaShape_2_visual():
    points = np.array(
        [
            [10, 10],
            [10, 14],
            [11, 11],
            [12, 11],
            [13, 10],
            [13, 13],
            [15, 14],
            [17, 16],
            [20, 20],
            [20, 24],
            [21, 21],
            [22, 21],
            [23, 20],
            [23, 23],
            [25, 24],
            [10, 24],
        ]
    )
    assert len(points) == 16

    alpha = 2.2
    alpha_shape = AlphaShape(alpha, points=points)
    H = alpha_shape.alpha_complex.graph_from_lines(alpha, type="regular")
    assert isinstance(H, nx.Graph)
    # visualization
    nx.draw_networkx(H, points, with_labels=True)
    plt.show()


def test_AlphaShape_2():
    points = np.array(
        [
            [10, 10],
            [10, 14],
            [11, 11],
            [12, 11],
            [13, 10],
            [13, 13],
            [15, 14],
            [17, 16],
            [20, 20],
            [20, 24],
            [21, 21],
            [22, 21],
            [23, 20],
            [23, 23],
            [25, 24],
            [10, 24],
        ]
    )
    assert len(points) == 16

    alpha = 2.2
    alpha_shape = AlphaShape(alpha, points=points)
    H = alpha_shape.alpha_complex.graph_from_lines(alpha, type="regular")
    assert isinstance(H, nx.Graph)

    assert len(alpha_shape.alpha_shape) == 25
    assert isinstance(alpha_shape.alpha_shape, list)
    assert alpha_shape.dimension == 2
    assert isinstance(alpha_shape.region, Region2D)
    assert alpha_shape.region_measure == 21.0
    assert alpha_shape.subregion_measure == 26.32455532033676
    assert alpha_shape.n_points_alpha_shape == 15
    assert alpha_shape.n_points_alpha_shape_rel == 0.9375
    assert alpha_shape.n_points_on_boundary == 10
    assert alpha_shape.n_points_on_boundary_rel == 0.6666666666666666
    assert np.array_equal(alpha_shape.vertex_indices, [0, 1, 2, 4, 5, 8, 9, 10, 12, 13])
    assert alpha_shape.vertices.shape == (10, 2)
    assert alpha_shape.vertices_alpha_shape.shape == (15, 2)
    assert len(alpha_shape.vertex_alpha_shape_indices) == 15
    cc = alpha_shape.connected_components
    assert len(cc) == 2
    assert isinstance(cc[0], Region2D)
    assert alpha_shape.vertices_connected_components_indices == [
        [13, 11, 12, 8, 10, 9],
        [2, 1, 0, 5, 3, 4],
    ]


def test_AlphaShape_3():
    points = np.array(
        [
            [10, 10],
            [10, 20],
            [20, 20],
            [20, 10],
            [10, 15],
            [15, 20],
            [20, 15],
            [15, 10],
            [10, 12],
            [10, 18],
            [20, 18],
            [20, 12],
            [17, 20],
            [12, 10],
            [17, 10],
            [12, 20],
            [11, 11],
            [11, 19],
            [19, 19],
            [19, 11],
            [11, 15],
            [15, 19],
            [19, 15],
            [15, 11],
            [13, 11],
            [11.5, 12],
            [12, 18],
            [13, 19],
            [17, 19],
            [18, 18],
            [17, 11],
            [18, 12],
            [15, 15],
        ]
    )
    assert len(points) == 33

    alpha = 2.2
    alpha_shape = AlphaShape(alpha, points=points)
    H = alpha_shape.alpha_complex.graph_from_lines(alpha, type="regular")
    assert isinstance(H, nx.Graph)
    assert len(alpha_shape.alpha_shape) == 75
    assert isinstance(alpha_shape.alpha_shape, list)
    assert alpha_shape.dimension == 2
    assert isinstance(alpha_shape.region, Region2D)
    assert alpha_shape.region_measure == 65.0
    assert alpha_shape.subregion_measure == 73.81471965135825
    assert alpha_shape.n_points_alpha_shape == 33
    assert alpha_shape.n_points_alpha_shape_rel == 1.0
    assert alpha_shape.n_points_on_boundary == 28
    assert alpha_shape.n_points_on_boundary_rel == 0.8484848484848485
    assert np.array_equal(
        alpha_shape.vertex_indices,
        [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
            12,
            13,
            14,
            15,
            20,
            21,
            22,
            23,
            24,
            25,
            26,
            27,
            28,
            29,
            30,
            31,
        ],
    )
    assert alpha_shape.vertices.shape == (28, 2)
    assert alpha_shape.vertices_alpha_shape.shape == (33, 2)
    assert len(alpha_shape.vertex_alpha_shape_indices) == 33
    cc = alpha_shape.connected_components
    assert len(cc) == 1
    assert isinstance(cc[0], Region2D)
    assert sorted(alpha_shape.vertices_connected_components_indices[0]) == sorted(
        [
            26,
            20,
            32,
            9,
            4,
            8,
            25,
            24,
            13,
            7,
            23,
            31,
            22,
            29,
            10,
            6,
            11,
            15,
            27,
            5,
            21,
            17,
            1,
            16,
            0,
            18,
            2,
            12,
            28,
            19,
            3,
            14,
            30,
        ]
    )



#!/usr/bin/env python

"""
Show original SMLM images overlaid with localization data.
Data is rendered in napari.

To run the script::

    locan check <pixel size> -f <images file> -l <localization file> -t <file type>

Try for instance::

    locan check 133 -f "locan/tests/test_data/images.tif" -l "locan/tests/test_data/rapidStorm_from_images.txt" -t 2
"""
import argparse

import numpy as np
import tifffile as tif

from locan import locdata_id
from locan.constants import FileType
from locan.dependencies import HAS_DEPENDENCY
from locan.gui import file_dialog
from locan.locan_io.locdata.io_locdata import load_locdata

if HAS_DEPENDENCY["napari"]:
    import napari


def render_locs_per_frame_napari(
    images,
    pixel_size,
    locdata,
    viewer=None,
    transpose=True,
    kwargs_image=None,
    kwargs_points=None,
):
    """
    Display original recording and overlay localization spots in napari.

    Parameters
    ---------
    images : np.array
        Stack of raw data as recorded by camera.
    pixel_size : float, tuple of float with shape (2,)
        Pixel size for images (in locdata units).
    transpose : bool
        If True transpose x and y axis of `images`.
    locdata : LocData
        Localization data that corresponds to `images` raw data.
    viewer : napari viewer
        The viewer object on which to add the image
    kwargs_image : dict
        Other parameters passed to napari.Viewer().add_image().
    kwargs_points : dict
        Other parameters passed to napari.Viewer().add_points().

    Returns
    -------
    napari Viewer object
        Viewer with the image.
    """
    if kwargs_image is None:
        kwargs_image = {}
    if kwargs_points is None:
        kwargs_points = {}

    if np.ndim(pixel_size) == 0:
        pixel_size_ = (pixel_size, pixel_size)
    elif np.ndim(pixel_size) == 1 and len(pixel_size) == 2:
        pixel_size_ = pixel_size
    else:
        raise TypeError("Dimension of `pixel_size` is incompatible with 2d image.")

    # transpose x and y axis
    if transpose:
        images_ = np.transpose(images, (0, 2, 1))
    else:
        images_ = images

    points = locdata.data[locdata.data["frame"] < len(images)][
        ["frame", "position_x", "position_y"]
    ].values

    # Provide napari viewer if not provided
    if viewer is None:
        viewer = napari.Viewer()

    viewer.add_image(images_, name="Raw data", **kwargs_image, scale=pixel_size_)
    viewer.add_points(
        data=points,
        name=f"LocData {locdata_id}",
        symbol="disc",
        size=500,
        face_color="r",
        edge_color="r",
        opacity=0.3,
        **kwargs_points,
    )

    return viewer


def sc_check(
    pixel_size,
    file_images=None,
    file_locdata=None,
    file_type=FileType.RAPIDSTORM,
    viewer=None,
    transpose=True,
    kwargs_image=None,
    kwargs_points=None,
):
    """
    Load and display original recording and load and overlay localization spots in napari.

    Parameters
    ---------
    pixel_size : float or tuple of float with shape (2,)
        Pixel size for images (in locdata units).
    file_images : str, bytes, os.PathLike
        File path for stack of raw data as recorded by camera.
    file_locdata : str, bytes, os.PathLike
        File path for localization data that corresponds to `images` raw data.
    file_type : int, str, locan.constants.FileType, locan.data.metadata_pb2.Metadata
        Indicator for the file type.
        Integer or string should be according to locan.constants.FileType.
    transpose : bool
        If True transpose x and y axis of `images`.
    locdata : LocData
        Localization data that corresponds to `images` raw data.
    viewer : napari.Viewer
        The viewer object on which to add the image
    kwargs_image : dict
        Other parameters passed to :meth:`napari.Viewer.add_image`.
    kwargs_points : dict
        Other parameters passed to :meth:`napari.Viewer.add_points`.

    Returns
    -------
    napar.Viewer
        Viewer with the image.
    """
    if kwargs_image is None:
        kwargs_image = {}
    if kwargs_points is None:
        kwargs_points = {}

    with napari.gui_qt():
        # load images
        if file_images is None:
            file_images = file_dialog(
                directory=None,
                message="Select images file...",
                filter="Tif files (*.tif)",
            )[0]
        # load images from tiff file
        image_stack = tif.imread(
            str(file_images)
        )  # , key=range(0, 10, 1))  - maybe add key parameter

        # load locdata
        if file_locdata is None:
            file_locdata = file_dialog(
                directory=None,
                message="Select a localization file...",
                filter="Text files (*.txt);; CSV files (*.csv)",
            )[0]
        locdata = load_locdata(file_locdata, file_type=file_type)

        # due to changed napari behavior from v3.0 on the context manager is moved up.
        # with napari.gui_qt():
        render_locs_per_frame_napari(
            image_stack,
            pixel_size,
            locdata,
            viewer,
            transpose,
            kwargs_image,
            kwargs_points,
        )


def _add_arguments(parser):
    parser.add_argument(
        dest="pixel_size", type=float, help="Pixel size for images (in locdata units)."
    )
    parser.add_argument(
        "-f",
        "--file",
        dest="file_images",
        type=str,
        default=None,
        help="File with images of original recording.",
    )
    parser.add_argument(
        "-l",
        "--localizations",
        dest="file_locdata",
        type=str,
        default=None,
        help="File with localization data.",
    )
    parser.add_argument(
        "-t",
        "--type",
        dest="file_type",
        type=int,
        default=2,
        help="Integer or string indicating the file type.",
    )


def main(args=None):

    parser = argparse.ArgumentParser(
        description="Show localizations in original recording."
    )
    _add_arguments(parser)
    returned_args = parser.parse_args(args)

    sc_check(
        pixel_size=returned_args.pixel_size,
        file_images=returned_args.file_images,
        file_locdata=returned_args.file_locdata,
        file_type=returned_args.file_type,
        transpose=True,
        kwargs_image={},
        kwargs_points={},
    )


if __name__ == "__main__":
    main()



import logging
import pickle
import tempfile
from io import StringIO
from pathlib import Path

import pytest

import locan.constants
from locan.data import metadata_pb2
from locan.dependencies import HAS_DEPENDENCY
from locan.locan_io import load_locdata, load_txt_file
from locan.locan_io.locdata.io_locdata import _map_file_type_to_load_function


def test_loading_txt_file(caplog):
    dat = load_txt_file(
        path=locan.ROOT_DIR / "tests/test_data/five_blobs.txt", nrows=10
    )
    # print(dat.data)
    assert len(dat) == 10

    dat = load_txt_file(
        path=locan.ROOT_DIR / "tests/test_data/five_blobs.txt",
        columns=["index", "position_x", "position_y", "cluster_label", "frame"],
        nrows=10,
    )
    # print(dat.data)
    assert len(dat) == 10

    file_like = StringIO("index,position_x,position_y,cluster_label\n" "0,624,919,3")
    dat = load_txt_file(
        path=file_like,
        columns=["index", "position_x", "position_y", "cluster_label"],
        nrows=1,
    )
    assert len(dat) == 1

    dat = load_txt_file(
        path=locan.ROOT_DIR / "tests/test_data/five_blobs.txt", columns=["c1"], nrows=10
    )
    # print(dat.data)
    assert caplog.record_tuples == [
        (
            "locan.locan_io.locdata.utilities",
            logging.WARNING,
            "Column c1 is not a Locan property standard.",
        )
    ]

    assert len(dat) == 10

    dat = load_txt_file(
        path=locan.ROOT_DIR / "tests/test_data/five_blobs.txt",
        columns=["c1"],
        nrows=10,
        property_mapping={"c1": "something"},
    )
    assert list(dat.data.columns) == ["something"]


def test__map_file_type_to_load_function():
    file_type = _map_file_type_to_load_function(file_type=1)
    assert callable(file_type)
    file_type = _map_file_type_to_load_function(file_type="RAPIDSTORM")
    assert callable(file_type)
    file_type = _map_file_type_to_load_function(
        file_type=locan.constants.FileType.RAPIDSTORM
    )
    assert callable(file_type)
    file_type = _map_file_type_to_load_function(file_type=metadata_pb2.RAPIDSTORM)
    assert callable(file_type)


def test_load_locdata():
    dat = load_locdata(
        path=locan.ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_data.txt",
        file_type="RAPIDSTORM",
        nrows=10,
    )
    assert len(dat) == 10

    dat = load_locdata(
        path=locan.ROOT_DIR / "tests/test_data/Elyra_dstorm_data.txt",
        file_type="ELYRA",
        nrows=10,
    )
    assert len(dat) == 10

    dat = load_locdata(
        path=locan.ROOT_DIR / "tests/test_data/Elyra_dstorm_data.txt",
        file_type="elyra",
        nrows=10,
    )
    assert len(dat) == 10

    dat = load_locdata(
        path=locan.ROOT_DIR / "tests/test_data/Elyra_dstorm_data.txt",
        file_type=3,
        nrows=10,
    )
    assert len(dat) == 10

    dat = load_locdata(
        path=locan.ROOT_DIR / "tests/test_data/Nanoimager_dstorm_data.csv",
        file_type="NANOIMAGER",
        nrows=10,
    )
    assert len(dat) == 10

    dat = load_locdata(
        path=locan.ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_track_data.txt",
        file_type="RAPIDSTORMTRACK",
        nrows=10,
    )
    assert len(dat) == 10

    dat = load_locdata(
        path=locan.ROOT_DIR / "tests/test_data/SMLM_dstorm_data.smlm",
        file_type="SMLM",
        nrows=10,
    )
    assert len(dat) == 10


@pytest.mark.skipif(not HAS_DEPENDENCY["h5py"], reason="requires h5py.")
def test_load_locdata_2():
    dat = load_locdata(
        path=locan.ROOT_DIR / "tests/test_data/decode_dstorm_data.h5",
        file_type="DECODE",
        nrows=10,
    )
    assert len(dat) == 10

    dat = load_locdata(
        path=locan.ROOT_DIR / "tests/test_data/smap_dstorm_data.mat",
        file_type="SMAP",
        nrows=10,
    )
    assert len(dat) == 10


def test_pickling_locdata(locdata_2d):
    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / "pickled_locdata.pickle"
        with open(file_path, "wb") as file:
            pickle.dump(locdata_2d, file, pickle.HIGHEST_PROTOCOL)
        with open(file_path, "rb") as file:
            locdata = pickle.load(file)
        assert len(locdata_2d) == len(locdata)
        assert isinstance(locdata.meta, metadata_pb2.Metadata)



import numpy as np
import pandas as pd
import pytest

from locan import LocData
from locan.analysis import LocalizationUncertaintyFromIntensity


@pytest.fixture()
def locdata_simple():
    locdata_dict = {
        "position_x": [0, 0, 1, 4, 5],
        "position_y": [0, 1, 3, 4, 1],
        "intensity": [0, 1, 3, 4, 1],
        "psf_sigma_x": [100, 100, 100, 100, 100],
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


def test_uncertainty_empty(caplog):
    LocalizationUncertaintyFromIntensity().compute(LocData())
    assert caplog.record_tuples == [
        ("locan.analysis.uncertainty", 30, "Locdata is empty.")
    ]


def test_uncertainty(locdata_simple):
    unc = LocalizationUncertaintyFromIntensity().compute(locdata_simple)
    # print(unc.results)
    # print(unc.results['Uncertainty_x'][0])
    assert unc.results["uncertainty_x"][0] == np.inf
    assert unc.results["uncertainty_x"][1] == 100



"""

Transform localization data.

This module takes localization data and applies transformation procedures on
coordinates or other properties.

"""
from __future__ import annotations

import logging
import sys

import numpy as np
import pandas as pd

from locan.data.locdata import LocData
from locan.data.metadata_utils import _modify_meta
from locan.data.region import Region
from locan.dependencies import HAS_DEPENDENCY
from locan.simulation import simulate_uniform

if HAS_DEPENDENCY["open3d"]:
    import open3d as o3d


__all__ = ["transform_affine", "randomize", "overlay"]

logger = logging.getLogger(__name__)


def _transform_affine_numpy(
    points, matrix=None, offset=None, pre_translation=None
) -> np.ndarray:
    """
    Transform `points` by an affine transformation using standard numpy
    procedures.

    Parameters
    ----------
    points : array-like
        Points on which to perform the manipulation.
    matrix : array-like | None
        Transformation matrix. If None the unit matrix is used.
        Array with shape (ndim, ndim).
        If None the unit matrix is used.
    offset : array-like | None
        Translation vector.
        Array with shape (ndim,).
        If None a vector of zeros is used.
    pre_translation : array-like | None
        Translation vector for coordinates applied before affine
        transformation. Array with shape (ndim,).
        The reverse translation is applied after the affine transformation.

    Returns
    -------
    numpy.ndarray
        Transformed coordinates.
    """
    points_ = np.asarray(points)
    dimension = np.shape(points_)[1]

    matrix_ = np.identity(dimension) if matrix is None else matrix
    offset_ = np.zeros(dimension) if offset is None else offset

    if pre_translation is None:
        # transformed_points = np.array(
        # [np.dot(matrix_, point) + offset_ for point in points_]
        # )
        # same function but better performance:
        transformed_points = np.einsum("ij, nj -> ni", matrix_, points) + offset_
    else:
        transformed_points = points_ + pre_translation
        # transformed_points = np.array(
        # [np.dot(matrix_, point) + offset_ for point in transformed_points]
        # )
        transformed_points = (
            np.einsum("ij, nj -> ni", matrix_, transformed_points) + offset_
        )
        transformed_points = transformed_points - pre_translation

    return transformed_points


def _homogeneous_matrix(matrix=np.identity(3), offset=np.zeros(3)) -> np.ndarray:
    """
    Combine transformation matrix and translation vector for dimension d into
    homogeneous (d+1, d+1) transformation
    matrix to be used with homogeneous coordinates.

    Parameters
    ----------
    matrix : array-like
        Transformation matrix. If None the unit matrix is used.
        Array with shape (ndim, ndim).
    offset : array-like | None
        Translation vector.
        Array with shape (ndim,).

    Returns
    -------
    numpy.ndarray
        Homogeneous transformation matrix to be used with homogeneous
        coordinate vector. Array with shape (ndim+1, ndim+1).
    """
    dimension = np.shape(matrix)[0]

    if dimension != np.shape(matrix)[1]:
        raise ValueError("The matrix has to be of shape (d, d).")

    if dimension != np.shape(offset)[0]:
        raise ValueError("The matrix and offset must have the same dimension.")

    matrix_ = np.identity(dimension + 1)
    matrix_[0:dimension, 0:dimension] = matrix
    matrix_[:dimension, dimension] = offset

    return matrix_


def _transform_affine_open3d(
    points, matrix=None, offset=None, pre_translation=None
) -> np.ndarray:
    """
    Transform `points` or coordinates in `locdata` by an affine
    transformation using open3d.

    Parameters
    ----------
    points : array-like
        Points on which to perform the manipulation.
    matrix : array-like | None
        Transformation matrix. If None the unit matrix is used.
        Array with shape (ndim, ndim).
        If None the unit matrix is used.
    offset : array-like | None
        Translation vector.
        Array with shape (ndim,).
        If None a vector of zeros is used.
    pre_translation : array-like | None
        Translation vector for coordinates applied before affine
        transformation. Array with shape (ndim,).
        The reverse translation is applied after the affine transformation.

    Returns
    -------
    numpy.ndarray
        Transformed coordinates.
    """
    if not HAS_DEPENDENCY["open3d"]:
        raise ImportError("open3d is required.")

    points_ = np.asarray(points)
    dimension = np.shape(points_)[1]

    matrix_ = np.identity(dimension) if matrix is None else matrix
    offset_ = np.zeros(dimension) if offset is None else offset

    # prepare 3d
    if dimension == 2:
        points_3d = np.concatenate([points_, np.zeros((len(points_), 1))], axis=1)
    elif dimension == 3:
        points_3d = points_
    else:
        raise ValueError("Point array has the wrong shape.")

    # points in open3d
    point_cloud = o3d.geometry.PointCloud()
    point_cloud.points = o3d.utility.Vector3dVector(points_3d)

    # transform
    matrix_3d = np.identity(3)
    matrix_3d[:dimension, :dimension] = matrix_
    offset_3d = np.zeros(3)
    offset_3d[:dimension] = offset_
    matrix_homogeneous = _homogeneous_matrix(matrix_3d, offset_3d)

    if pre_translation is None:
        point_cloud.transform(matrix_homogeneous)
    else:
        pre_translation_3d = np.zeros(3)
        pre_translation_3d[:dimension] = pre_translation
        point_cloud.translate(pre_translation_3d)
        point_cloud.transform(matrix_homogeneous)
        point_cloud.translate(-pre_translation_3d)

    # return correct dimension
    if dimension == 2:
        transformed_points = np.asarray(point_cloud.points)[:, 0:2]
    else:  # if dimension == 3:
        transformed_points = np.asarray(point_cloud.points)

    return transformed_points


def transform_affine(
    locdata, matrix=None, offset=None, pre_translation=None, method="numpy"
) -> np.ndarray | LocData:
    """
    Transform `points` or coordinates in `locdata` by an affine transformation.

    Parameters
    ----------
    locdata : numpy.ndarray | LocData
        Localization data on which to perform the manipulation.
    matrix : array-like | None
        Transformation matrix. If None the unit matrix is used.
        Array with shape (ndim, ndim).
        If None the unit matrix is used.
    offset : array-like | None
        Translation vector.
        Array with shape (ndim,).
        If None a vector of zeros is used.
    pre_translation : array-like | None
        Translation vector for coordinates applied before affine
        transformation. Array with shape (ndim,).
        The reverse translation is applied after the affine transformation.
    method : str
        The method (i.e. library or algorithm) used for computation.
        One of 'numpy', 'open3d'.

    Returns
    -------
    numpy.ndarray | LocData
        New localization data with transformed coordinates.
    """
    local_parameter = locals()

    if len(locdata) == 0:
        return locdata

    # adjust input
    if isinstance(locdata, LocData):
        points = locdata.coordinates
    else:
        points = locdata

    if method == "numpy":
        transformed_points = _transform_affine_numpy(
            points, matrix=matrix, offset=offset, pre_translation=pre_translation
        )
    elif method == "open3d":
        transformed_points = _transform_affine_open3d(
            points, matrix=matrix, offset=offset, pre_translation=pre_translation
        )
    else:
        raise ValueError(f"Method {method} is not available.")

    # prepare output
    if isinstance(locdata, LocData):
        # new LocData object
        new_dataframe = locdata.data.copy()
        new_dataframe.update(
            pd.DataFrame(
                transformed_points,
                columns=locdata.coordinate_labels,
                index=locdata.data.index,
            )
        )
        new_locdata = LocData.from_dataframe(new_dataframe)

        # update metadata
        meta_ = _modify_meta(
            locdata,
            new_locdata,
            function_name=sys._getframe().f_code.co_name,
            parameter=local_parameter,
            meta=None,
        )
        new_locdata.meta = meta_

        return new_locdata

    else:
        return transformed_points


def randomize(locdata, hull_region="bb", seed=None):
    """
    Transform locdata coordinates into randomized coordinates that follow
    complete spatial randomness on the same region as the input locdata.

    Parameters
    ----------
    locdata : LocData
        Localization data to be randomized
    hull_region : Region | str
        Region of interest. String identifier can be one of 'bb', 'ch', 'as',
        'obb' referring to the corresponding hull.
    seed : None, int, array_like[int], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    locdata : LocData
        New localization data with randomized coordinates.
    """
    local_parameter = locals()

    rng = np.random.default_rng(seed)

    if hull_region == "bb":
        region_ = locdata.bounding_box.hull.T
    elif hull_region == "ch":
        region_ = locdata.convex_hull.region
    elif hull_region == "as":
        region_ = locdata.alpha_shape.region
    elif hull_region == "obb":
        region_ = locdata.oriented_bounding_box.region
    elif isinstance(hull_region, Region):
        region_ = hull_region
    else:
        raise NotImplementedError

    new_locdata = simulate_uniform(n_samples=len(locdata), region=region_, seed=rng)

    # update metadata
    meta_ = _modify_meta(
        locdata,
        new_locdata,
        function_name=sys._getframe().f_code.co_name,
        parameter=local_parameter,
        meta=None,
    )
    new_locdata.meta = meta_

    return new_locdata


def overlay(locdatas, centers="centroid", orientations=None):
    """
    Translate locdatas to their common center and rotate according to their
    orientation.

    Parameters
    ----------
    locdatas : Iterable[LocData]
        Localization data to overlay.
    centers : Iterable | str | None
        centers to which locdatas are translated.
        Must have the same length as locdatas.
        One of `centroid`, `ch`, 'bb', 'obb', or 'region'.
        If None, no translation is applied.
    orientations : Iterable | str | None
        Orientation value to use in degree.
        Must have the same length as locdatas.
        If str, it must be one of `orientation_im`, `orientation_obb`.
        If None, no rotation is applied.

    Returns
    -------
    LocData
        Collection with transformed locdatas.
    """
    local_parameter = locals()

    if not locdatas or not isinstance(locdatas, (tuple, list)):
        raise TypeError("locdatas must be a list of LocData objects.")

    dimensions = {locdata.dimension for locdata in locdatas}
    if len(dimensions) == 1:
        dimension = dimensions.pop()
    else:
        raise ValueError(
            "Some elements of `locdatas` have a different dimension. "
            "All locdatas must have the same dimension."
        )

    if centers is None or isinstance(centers, str):
        centers = [centers] * len(locdatas)
    else:
        if len(centers) != len(locdatas):
            raise ValueError("Centers must have the same length as locdatas.")

    if orientations is None or isinstance(orientations, str):
        orientations = [orientations] * len(locdatas)
    else:
        if len(orientations) != len(locdatas):
            raise ValueError("orientations must have the same length as locdatas.")

    transformed_locdatas = []
    for locdata, centre, orientation in zip(locdatas, centers, orientations):
        # translation
        if centre is None:
            transformed_locdata = locdata
        elif not isinstance(centre, str):
            transformed_locdata = transform_affine(
                locdata, matrix=None, offset=np.multiply(centre, -1)
            )
        elif centre == "centroid":
            transformed_locdata = transform_affine(
                locdata, matrix=None, offset=np.multiply(locdata.centroid, -1)
            )
        elif centre == "bb":
            transformed_locdata = transform_affine(
                locdata,
                matrix=None,
                offset=np.multiply(locdata.bounding_box.region.centroid, -1),
            )
        elif centre == "obb":
            transformed_locdata = transform_affine(
                locdata,
                matrix=None,
                offset=np.multiply(locdata.oriented_bounding_box.region.centroid, -1),
            )
        elif centre == "ch":
            transformed_locdata = transform_affine(
                locdata,
                matrix=None,
                offset=np.multiply(locdata.convex_hull.region.centroid, -1),
            )
        elif centre == "region":
            transformed_locdata = transform_affine(
                locdata, matrix=None, offset=np.multiply(locdata.region.centroid, -1)
            )
        else:
            raise ValueError(f"Value for centre={centre} is not defined.")

        # prepare rotation
        if orientation is None:
            matrix = np.identity(dimension)
        elif not isinstance(orientation, str):
            theta = -np.radians(orientation)
            cos_, sin_ = np.cos(theta), np.sin(theta)
            matrix = np.array(((cos_, -sin_), (sin_, cos_)))
        else:
            if dimension != 2:
                raise NotImplementedError(
                    "Rotation has only been implemented for 2 dimensions."
                )
            if orientation == "orientation_obb":
                locdata.oriented_bounding_box
            elif orientation == "orientation_im":
                locdata.inertia_moments
            else:
                raise ValueError(f"orientation={orientation} is undefined.")

            theta = -np.radians(locdata.properties[orientation])
            cos_, sin_ = np.cos(theta), np.sin(theta)
            matrix = np.array(((cos_, -sin_), (sin_, cos_)))

        # rotation
        transformed_locdata = transform_affine(
            transformed_locdata, matrix=matrix, offset=None
        )

        transformed_locdatas.append(transformed_locdata)

    new_locdata = LocData.from_collection(transformed_locdatas)

    # update metadata
    del new_locdata.meta.history[:]
    new_locdata.meta.history.add(
        name=sys._getframe().f_code.co_name, parameter=str(local_parameter)
    )

    return new_locdata



"""

File input/output for localization data in rapidSTORM files.

"""
import logging
import xml.etree.ElementTree as etree

import numpy as np
import pandas as pd

import locan.constants
from locan.data import metadata_pb2
from locan.data.locdata import LocData
from locan.locan_io.locdata.utilities import (
    convert_property_names,
    convert_property_types,
    open_path_or_file_like,
)

__all__ = [
    "load_rapidSTORM_header",
    "load_rapidSTORM_file",
    "load_rapidSTORM_track_header",
    "load_rapidSTORM_track_file",
]

logger = logging.getLogger(__name__)


def _read_rapidSTORM_header(file):
    """
    Read xml header from a rapidSTORM single-molecule localization file and identify column names.

    Parameters
    ----------
    file : file-like
        A rapidSTORM file to load.

    Returns
    -------
    list of str
        A list of valid dataset property keys as derived from the rapidSTORM identifiers.
    """
    # read xml part in header
    header = file.readline()
    header = header[2:]

    # get iteratible
    parsed = etree.XML(header)

    # list identifiers
    identifiers = []
    for elem in parsed:
        for name, value in sorted(elem.attrib.items()):
            if name == "identifier":
                identifiers.append(value)

    # turn identifiers into valuable LocData keys
    column_keys = convert_property_names(
        properties=identifiers, property_mapping=locan.constants.RAPIDSTORM_KEYS
    )
    return column_keys


def load_rapidSTORM_header(path):
    """
    Load xml header from a rapidSTORM single-molecule localization file and identify column names.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a rapidSTORM file to load.

    Returns
    -------
    list of str
        A list of valid dataset property keys as derived from the rapidSTORM identifiers.
    """

    # read xml part in header
    with open_path_or_file_like(path) as file:
        return _read_rapidSTORM_header(file)


def load_rapidSTORM_file(path, nrows=None, convert=True, **kwargs):
    """
    Load data from a rapidSTORM single-molecule localization file.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a rapidSTORM file to load.
    nrows : int, None
        The number of localizations to load from file. None means that all available rows are loaded.
    convert : bool
        If True convert types by applying type specifications in locan.constants.PROPERTY_KEYS.
    kwargs : dict
        Other parameters passed to `pandas.read_csv()`.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    with open_path_or_file_like(path) as file:
        columns = _read_rapidSTORM_header(file)
        dataframe = pd.read_csv(
            file, sep=" ", skiprows=0, nrows=nrows, names=columns, **kwargs
        )

    if convert:
        dataframe = convert_property_types(
            dataframe, types=locan.constants.PROPERTY_KEYS
        )

    dat = LocData.from_dataframe(dataframe=dataframe)

    dat.meta.source = metadata_pb2.EXPERIMENT
    dat.meta.state = metadata_pb2.RAW
    dat.meta.file.type = metadata_pb2.RAPIDSTORM
    dat.meta.file.path = str(path)

    for property in sorted(
        list(set(columns).intersection({"position_x", "position_y", "position_z"}))
    ):
        dat.meta.localization_properties.add(name=property, unit="nm", type="float")

    del dat.meta.history[:]
    dat.meta.history.add(
        name="load_rapidSTORM_file", parameter="path={}, nrows={}".format(path, nrows)
    )

    return dat


def _read_rapidSTORM_track_header(file):
    """
    Read xml header from a rapidSTORM (track) single-molecule localization file and identify column names.

    Parameters
    ----------
    file : file-like
        A rapidSTORM file to load.

    Returns
    -------
    list of str
        A list of valid dataset property keys as derived from the rapidSTORM identifiers.
    """
    # read xml part in header
    header = file.readline()
    header = header[2:]

    parsed = etree.XML(header)

    # list identifiers
    identifiers = []
    for field in parsed.findall("field"):
        identifiers.append(field.get("identifier"))

    # turn identifiers into valuable LocData keys
    column_keys = convert_property_names(
        properties=identifiers, property_mapping=locan.constants.RAPIDSTORM_KEYS
    )

    # list child identifiers
    child_identifiers = []
    for field in parsed.findall("localizations"):
        for field_ in field.findall("field"):
            child_identifiers.append(field_.get("identifier"))

    # turn child identifiers into valuable LocData keys
    column_keys_tracks = convert_property_names(
        properties=child_identifiers, property_mapping=locan.constants.RAPIDSTORM_KEYS
    )

    return column_keys, column_keys_tracks


def load_rapidSTORM_track_header(path):
    """
    Load xml header from a rapidSTORM (track) single-molecule localization file and identify column names.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a rapidSTORM file to load.

    Returns
    -------
    list of str
        A list of valid dataset property keys as derived from the rapidSTORM identifiers.
    """

    # read xml part in header
    with open_path_or_file_like(path) as file:
        return _read_rapidSTORM_track_header(file)


def load_rapidSTORM_track_file(
    path, nrows=None, convert=True, collection=True, min_localization_count=1, **kwargs
):
    """
    Load data from a rapidSTORM single-molecule localization file with tracked localizations.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a rapidSTORM file to load.
    nrows : int, None
        The number of localizations to load from file. None means that all available rows are loaded.
    convert : bool
        If True convert types by applying type specifications in locan.constants.PROPERTY_KEYS.
    collection : bool
        If True a collection of all tracks is returned. If False LocData with center positions is returned.
    min_localization_count : int
        If collection is True, only clusters with at least `min_localization_count` localizations are loaded.
    kwargs : dict
        Other parameters passed to `pandas.read_csv()`.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations/tracks as a collection.
    """
    with open_path_or_file_like(path) as file:
        columns, columns_track = _read_rapidSTORM_track_header(file)
        lines = pd.read_csv(
            file, lineterminator="\n", nrows=nrows, skiprows=1, header=None, **kwargs
        )

    lines = lines[0].str.split(" ", expand=False)

    if collection:
        # prepare dataframes with tracked localizations
        tracks = [
            np.array(line[len(columns) + 1 :]).reshape(-1, len(columns_track))
            for line in lines
            if int(line[len(columns)]) >= min_localization_count
        ]
        # +1 to account for the column with number of locs in track
        track_list = []
        for track in tracks:
            dataframe = pd.DataFrame(track, columns=columns_track)
            if convert:
                dataframe = convert_property_types(
                    dataframe, types=locan.constants.PROPERTY_KEYS
                )
            else:
                dataframe = dataframe.convert_dtypes()
            dat = LocData.from_dataframe(dataframe=dataframe)
            track_list.append(dat)

        collection = LocData.from_collection(track_list)

        collection.meta.source = metadata_pb2.EXPERIMENT
        collection.meta.state = metadata_pb2.RAW
        collection.meta.file.type = metadata_pb2.RAPIDSTORMTRACK
        collection.meta.file.path = str(path)

        for property in sorted(
            list(
                set(columns_track).intersection(
                    {"position_x", "position_y", "position_z"}
                )
            )
        ):
            collection.meta.localization_properties.add(
                name=property, unit="nm", type="float"
            )

        del collection.meta.history[:]
        collection.meta.history.add(
            name="load_rapidSTORM_track_file",
            parameter="path={}, nrows={}".format(path, nrows),
        )

        return collection

    else:
        # prepare dataframe with center track positions
        dataframe = pd.DataFrame(
            [line[: len(columns)] for line in lines], columns=columns
        )
        if convert:
            dataframe = convert_property_types(
                dataframe, types=locan.constants.PROPERTY_KEYS
            )
        else:
            dataframe = dataframe.convert_dtypes()

        locdata = LocData.from_dataframe(dataframe=dataframe)

        locdata.meta.source = metadata_pb2.EXPERIMENT
        locdata.meta.state = metadata_pb2.RAW
        locdata.meta.file.type = metadata_pb2.RAPIDSTORM
        locdata.meta.file.path = str(path)

        for property in sorted(
            list(set(columns).intersection({"position_x", "position_y", "position_z"}))
        ):
            locdata.meta.localization_properties.add(
                name=property, unit="nm", type="float"
            )

        del locdata.meta.history[:]
        locdata.meta.history.add(
            name="load_rapidSTORM_track_file",
            parameter="path={}, nrows={}".format(path, nrows),
        )

        return locdata



import pandas as pd
from scipy import stats

from locan import LocData
from locan.analysis import LocalizationProperty
from locan.analysis.localization_property import _DistributionFits


def test_test_LocalizationProperty_empty(caplog):
    lprop = LocalizationProperty().compute(LocData())
    lprop.fit_distributions()
    lprop.plot()
    lprop.hist()
    assert caplog.record_tuples == [
        ("locan.analysis.localization_property", 30, "Locdata is empty."),
        ("locan.analysis.localization_property", 30, "No results available to fit."),
    ]


def test_DistributionFits(locdata_rapidSTORM_2d):
    lprop = LocalizationProperty().compute(locdata=locdata_rapidSTORM_2d)
    assert repr(lprop) == "LocalizationProperty(loc_property=intensity, index=None)"
    ds = _DistributionFits(lprop)
    ds.fit(distribution=stats.expon)
    # print(ds.parameters)
    # print(ds.__dict__)
    # #print(ds.parameter_dict())
    # ds.plot()
    assert ds.parameters == ["intensity_loc", "intensity_scale"]
    ds.fit(distribution=stats.expon, with_constraints=False, floc=1000)
    assert ds.parameter_dict()["intensity_loc"] == 1000


def test_Localization_property(locdata_rapidSTORM_2d):
    lprop = LocalizationProperty(loc_property="intensity")
    assert not lprop
    lprop.plot()
    lprop.hist()
    lprop.compute(locdata=locdata_rapidSTORM_2d)
    assert lprop
    # print(lprop)
    assert repr(lprop) == "LocalizationProperty(loc_property=intensity, index=None)"
    assert lprop.results.columns == pd.Index(["intensity"], dtype="object")
    assert lprop.distribution_statistics is None
    lprop.plot()
    lprop.hist()
    lprop.fit_distributions()
    assert list(lprop.distribution_statistics.parameter_dict().keys()) == [
        "intensity_loc",
        "intensity_scale",
    ]

    lprop = LocalizationProperty(loc_property="intensity", index="frame").compute(
        locdata=locdata_rapidSTORM_2d
    )
    # print(lprop.results)
    assert lprop.results.index.name == "frame"
    assert lprop.results.columns == pd.Index(["intensity"], dtype="object")



"""

Utility functions for interacting with napari.

"""
from __future__ import annotations

import logging
from pathlib import Path
from typing import Iterable  # noqa: F401

import numpy as np

from locan.data.region import Ellipse, Polygon, Rectangle, Region
from locan.data.rois import Roi
from locan.dependencies import HAS_DEPENDENCY
from locan.gui import file_dialog
from locan.visualize.napari.render2d import render_2d_napari

if HAS_DEPENDENCY["napari"]:
    import napari

logger = logging.getLogger(__name__)

__all__ = [
    "select_by_drawing_napari",
    "get_rois",
    "save_rois",
]


def select_by_drawing_napari(locdata, napari_run=True, **kwargs):
    """
    Select region of interest from rendered image by drawing rois in napari.

    Rois will be created from shapes in napari.viewer.layers['Shapes'].

    Parameters
    ----------
    locdata : LocData
        The localization data from which to select localization data.
    napari_run : bool
        If `True` napari.run is called (set to `False` for testing).
    kwargs : dict
        Other parameters passed to :func:`render_2d_napari`.

    Returns
    -------
    list[Roi]

    See Also
    --------
    :func:`locan.scripts.rois` : script for drawing rois
    """
    # select roi
    viewer = render_2d_napari(locdata, **kwargs)
    if "Rois" not in viewer.layers:
        viewer.add_shapes(name="Rois", edge_width=0.1)
    if napari_run:
        napari.run()

    roi_list = get_rois(viewer.layers["Rois"], reference=locdata)

    return roi_list


def _shape_to_region(vertices, shape_type):
    """
    Convert napari shape to `locan.Region`.

    Parameters
    ----------
    vertices : np.array[float]
        Sequence of point coordinates as returned by napari.
    shape_type : str
        One of rectangle, ellipse, or polygon.

    Returns
    -------
    Region
    """
    if shape_type == "rectangle":
        if len(set(vertices[:, 0].astype(int))) != 2:
            raise NotImplementedError("Rotated rectangles are not implemented.")
        mins = vertices.min(axis=0)
        maxs = vertices.max(axis=0)
        corner_x, corner_y = mins
        width, height = maxs - mins
        angle = 0
        region = Rectangle((corner_x, corner_y), width, height, angle)

    elif shape_type == "ellipse":
        if len(set(vertices[:, 0].astype(int))) != 2:
            raise NotImplementedError("Rotated ellipses are not implemented.")
        mins = vertices.min(axis=0)
        maxs = vertices.max(axis=0)
        width, height = maxs - mins
        center_x, center_y = mins[0] + width / 2, mins[1] + height / 2
        angle = 0
        region = Ellipse((center_x, center_y), width, height, angle)

    elif shape_type == "polygon":
        region = Polygon(np.concatenate([vertices, [vertices[0]]], axis=0))

    else:
        raise TypeError(f" Type {shape_type} is not defined.")

    return region


def _shapes_to_regions(shapes_data) -> list[Region]:
    """
    Convert napari shapes to `locan.Region`.

    Parameters
    ----------
    shapes_data : napari.types.ShapesData
        Shapes data with list of shapes

    Returns
    -------
    list[Region]
    """
    if shapes_data[2] != "shapes":
        raise ValueError("shapes_data[2] must equal 'shapes'.")
    data = shapes_data[0]
    shape_types = shapes_data[1]["shape_type"]
    regions = [
        _shape_to_region(vertices=vertices, shape_type=shape_type)
        for vertices, shape_type in zip(data, shape_types)
    ]
    return regions


def get_rois(shapes_layer, reference=None, loc_properties=None):
    """
    Create rois from shapes in napari.viewer.Shapes.

    Parameters
    ----------
    shapes_layer: napari.layers.Shapes
        Napari shapes layer like `viewer.layers["Shapes"]`
    reference : LocData, dict, locan.data.metadata_pb2.Metadata, locan.data.metadata_pb2.File, None
        Reference to localization data for which the region of interest
        is defined. It can be a LocData object, a reference to a saved
        SMLM file, or None for indicating no specific reference.
        When dict it must have keys `file_path`and `file_type`.
        When Metadata message it must have keys `file.path` and
        `file.type` for a path pointing to a localization file and an
        integer or string indicating the file type.
        Integer or string should be according to
        locan.constants.FileType.
    loc_properties : tuple of str
        Localization properties in LocData object on which the region
        selection will be applied (for instance the coordinate_labels).

    Returns
    -------
    list[Roi]

    See Also
    --------
    :func:`locan.scripts.rois` : script for drawing rois
    """
    shapes_data = shapes_layer.as_layer_data_tuple()
    regions = _shapes_to_regions(shapes_data=shapes_data)
    rois = [
        Roi(region=reg, reference=reference, loc_properties=loc_properties)
        for reg in regions
    ]
    return rois


def save_rois(rois, file_path=None, roi_file_indicator="_roi"):
    """
    Save list of Roi objects.

    Parameters
    ----------
    rois : Iterable[Rois]
        The rois to be saved.
    file_path : str | bytes | os.PathLike | "roi_reference" | None
        Base name for roi files or existing directory to save rois in.
        If "roi_reference", roi.reference.file.path is used.
        If None, a file dialog is opened.
    roi_file_indicator : str
        Indicator to add to the localization file name and use as roi
        file name (with further extension .yaml).

    Returns
    -------
    list[str]
        New created roi file paths.
    """
    # choose file interactively
    if file_path is None:
        file_path = Path(
            file_dialog(message="choose localization file or base name")[0]
        )
    elif file_path == "roi_reference":
        file_path = None
    elif Path(file_path).is_dir():
        file_path = Path(file_path) / "my"
    else:
        file_path = Path(file_path)

    # create roi file names and save rois
    roi_path_list = []
    for i, roi in enumerate(rois):
        if file_path is None:
            try:
                file_path = Path(roi.reference.file.path)
            except AttributeError:
                raise
        roi_file = file_path.stem + roi_file_indicator + f"_{i}.yaml"
        roi_path = file_path.with_name(roi_file)
        roi_path_list.append(roi_path)
        roi.to_yaml(path=roi_path)

    return roi_path_list
